import numpy as np

from lentil import util
from lentil import zernike

__all__ = ['power_spectrum', 'translation_defocus']


def power_spectrum(mask, pixelscale, rms, half_power_freq, exp, seed=None):
    """Wavefront error defined by a Power Spectral Density (PSD) function.

    Parameters
    ----------
    mask : array_like
        Binary [0,1] mask defining pupil extent

    pixelscale : float
        Physical size of each pixel in the resulting opd in meters.

    rms : float
        RMS value of the PSD error in meters

    half_power_freq : float
        Half-power frequency in number of cycles per pixel

    exp : float
        Exponent of the inverse-power law

    seed : int, optional
        Random seed used to initialize the pseudo-random number generator. If
        seed is `None` (default), the seed will be randomly generated from 
        ``/dev/urandom`` if available or the system clock.

    Returns
    -------
    wfe : ndarray
        Masked wavefront error with requested PSD

    References
    ----------
    [1] Sidick (2009) Power Spectral Density Specification and Analysis of
    Large Optical Surfaces

    """

    mask = np.asarray(mask)
    rng = np.random.RandomState(seed)

    # Define a frequency grid in units of cycles/px
    n, m = mask.shape
    yy, xx = np.mgrid[0:m, 0:n]
    yy = (yy - (np.floor(m / 2) + 1)) / m
    xx = (xx - (np.floor(n / 2) + 1)) / n
    dr = np.sqrt(xx * xx + yy * yy)

    # Scale the half-power point in units of cycle/px
    half_power_freq = half_power_freq * pixelscale / np.sqrt(m ** 2 + n ** 2)

    # Generate the PSD function and noise filter
    psd = 1 / (1 + (dr / half_power_freq) ** exp)
    psd[dr == 0] = 0
    psd = psd / np.sum(psd)

    H = np.fft.fftshift(np.sqrt(psd))

    n, m = mask.shape

    # Generate noise, filter it to realize the requested PSD, and enforce
    # the pupil mask
    noise = rng.normal(size=[n, m])
    phase = np.real(np.fft.ifft2(np.fft.fft2(noise) * H)) * np.sqrt(m * n)

    phase *= mask
    phase = phase * np.sqrt(np.count_nonzero(phase)/np.sum(np.abs(phase)**2)) * rms

    return phase


def translation_defocus(mask, f_number, translation):
    """Defocus error characterized by a axial translation.

    Parameters
    ----------
    mask : array_like
        Binary [0,1] mask defining the pupil extent to compute the wavefront
        over.

    f_number : float
        System or beam f-number.

    translation : float
        Axial translation from best focus given in the same units as the
        resulting OPD.

    Returns
    -------
    wfe : ndarray
        Focus error due to axial translation

    Warning
    -------
    Defocus is computed over a circular pupal defined by the maximum
    horizontal or vertical extent of the supplied mask before applying the
    mask. As a result, the RMS and peak-to-valley wavefront measurements
    of the returned OPD will be different than what was requested for non-
    circular masks.

    """
    mask = np.asarray(mask)

    rmin, rmax, cmin, cmax = util.boundary(mask)
    extent = max(rmax-rmin, cmax-cmin)

    circlemask = util.circle(mask.shape, np.ceil(extent/2))

    coeff = translation/(8*f_number**2)  # p-v defocus

    z4 = zernike.zernike(circlemask, 4)
    z4 = z4/(z4.max() - z4.min())  # normalize p-v = 1

    return z4*mask*coeff


# def kolmogorov_wfe():
#     """Wavefront error representing a turbulent phase screen."""
#     pass



from math import factorial

import numpy as np

from lentil import util

__all__ = ['zernike', 'zernike_compose', 'zernike_fit', 'zernike_remove',
           'zernike_basis', 'zernike_coordinates']


def zernike(mask, index, normalize=True, rho=None, theta=None):
    """Compute the circular Zernike polynomial for a given mask.

    Parameters
    ----------
    mask : array_like
        Binary mask defining the extent to compute the Zernike polynomial
        over.

    index : int
        Noll Zernike index as defined in [1]

    normalize : bool, optional
        If True (default), the output is normalized according to [1]. If False,
        the output value ranges [-1, 1] over the mask.

    rho : array_like, optional
        Radial coordinates of the mask array. :attr:`rho` should be 0 at the
        origin and 1 at the edge of the circle.

    theta : array_like, optional
        Angular coordinates of the mask array in radians.

    Returns
    -------
    out
        Circular Zernike polynomial computed over the given mask.

    Warning
    -------
    Zernike polynomials are defined to be orthogonal on the unit circle. If
    the supplied mask is non-circular, the Zernike polynomial is computed on an
    outscribing circle and then cropped by the mask. Note that this operation
    breaks the orthogonality of the Zernike polynomial. When working with a
    non-circular mask, care must be taken to understand any side-effects of
    using Zernike polynomials constructed in this manner. Any undesirable
    side-effects can be mitigated by using Zernike polynomials that have
    undergone Gram-Schmidt orthogonalization over the supplied mask.

    References
    ----------
    [1] Noll, RJ. Zernike polynomials and atmospheric turbulence. J Opt Soc Am 66, 207-211  (1976).

    """
    mask = np.asarray(mask)
    #out = np.zeros_like(mask)
    #mask_slice = util.boundary_slice(mask, pad=0)
    #mask = mask[mask_slice]
    
    if rho is None:
        rho, theta = zernike_coordinates(mask)
    else:
        if theta is None:
            raise ValueError("Both rho and theta must be specified")

    m, n = zernike_index(index)

    if m == 0:
        if n == 0:
            Z = mask
        else:
            if normalize:
                Z = np.sqrt(n+1) * R(m, n, rho) * mask
            else:
                Z = R(m, n, rho) * mask

    elif m > 0:
        if normalize:
            Z = np.sqrt(2) * np.sqrt(n+1) * R(m, n, rho) * np.cos(m*theta) * mask
        else:
            Z = R(m, n, rho) * np.cos(m*theta) * mask

    else:
        if normalize:
            Z = np.sqrt(2) * np.sqrt(n+1) * R(m, n, rho) * np.sin(m*theta) * mask
        else:
            Z = R(m, n, rho) * np.sin(m*theta) * mask

    #out[mask_slice] = Z
    out = Z
    return out


def R(m, n, rho):

    m = int(np.abs(m))
    n = int(np.abs(n))

    if (n - m) & 1:  # odd
        return 0
    else:
        R = np.zeros(rho.shape)
        for k in range(int(n-m)//2 + 1):
            Rk = ((-1) ** k * factorial(n-k) /
                  (factorial(k) * factorial((n+m)/2-k) * factorial((n-m)/2-k)))
            R += Rk * rho ** (n-2*k)
        return R


def zernike_compose(mask, coeffs, normalize=True, rho=None, theta=None):
    """Create an OPD based on the supplied Zernike coefficients.

    Parameters
    ----------
    mask : array_like
        Binary mask defining the extent to compute the Zernike polynomial over.

    coeffs : array_like
        List of coefficients corresponding to Zernike indices (Noll ordering)
        used to create the OPD.

    normalize : bool, optional
        If True (default), the output is normalized according to [1]. If False,
        the output value ranges [-1, 1] over the mask.

    rho : array_like, optional
        Radial coordinates of the mask array. :attr:`rho` should be 0 at the
        origin and 1 at the edge of the circle.

    theta : array_like, optional
        Angular coordinates of the mask array in radians.

    Returns
    -------
    ndarray
        OPD

    Examples
    --------
    Compute a random OPD using the first ten Zernikes:

    .. code:: pycon

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> import lentil
        >>> mask = lentil.util.circlemask((256,256), 128)
        >>> coeffs = np.random.rand(10)*1e-8
        >>> opd = lentil.zernike.zernike_compose(mask, coeffs)
        >>> plt.imshow(opd)

    .. image:: /_static/img/api/zernike/zernike_compose_1.png
        :width: 300px

    Using the same mask, compute an OPD representing 200 nm focus error (Z4) and -100 nm
    astigmatism error (Z6):

    .. code:: pycon

        >>> opd = mo.zernike.zernike_compose(mask, [0, 0, 0, 200e-9, 0, -100e-9])
        >>> plt.imshow(opd)

    .. image:: /_static/img/api/zernike/zernike_compose_2.png
        :width: 300px

    References
    ----------
    [1] Noll, RJ. Zernike polynomials and atmospheric turbulence. J Opt Soc Am 66, 207-211  (1976).

    """
    mask = np.asarray(mask)
    coeffs = np.asarray(coeffs)
    opd = np.zeros(mask.shape)

    for index, coeff in np.ndenumerate(coeffs):
        opd += coeff * zernike(mask, index[0]+1, normalize, rho, theta)

    return opd


def zernike_basis(mask, modes, vectorize=False, normalize=True, rho=None, theta=None):
    """Compute a Zernike basis set for a given mask.

    Parameters
    ----------
    mask : array_like
        Binary mask defining the extent to compute the Zernike polynomial over.

    modes : array_like
        List of modes (Noll ordering) to return.

    vectorize : bool, optional
        If True, the output is returned as a
        ``(length(modes), modes.shape[0]*modes.shape[1])`` array If False
        (default), the output is returned as a
        ``(length(terms), mask.shape[0], mask.shape[1])`` cube.

    normalize : bool, optional
        If True (default), the output is normalized according to [1]. If False,
        the output value ranges [-1, 1] over the mask.

    rho : array_like, optional
        Radial coordinates of the mask array. :attr:`rho` should be 0 at the
        origin and 1 at the edge of the circle.

    theta : array_like, optional
        Angular coordinates of the mask array in radians.

    Returns
    -------
    ndarray
        Zernike basis set

    References
    ----------
    [1] Noll, RJ. Zernike polynomials and atmospheric turbulence. J Opt Soc Am 66, 207-211  (1976).

    """
    modes = np.asarray(modes)
    if modes.shape == ():
        modes = modes[..., np.newaxis]

    mask = np.asarray(mask)
    basis = np.zeros(modes.shape + mask.shape)

    for index, mode in np.ndenumerate(modes):
        basis[index] = zernike(mask, mode, normalize, rho, theta)

    if vectorize:
        # reshape basis from cube to matrix
        return basis.reshape(basis.shape[0], -1)
    else:
        return basis


def zernike_fit(opd, mask, modes, normalize=True, rho=None, theta=None):
    """Fit a Zernike basis set to an OPD.

    Parameters
    ----------
    opd : array_like
        OPD to fit.

    mask : array_like
        Binary mask defining the extent to compute the Zernike basis over.

    modes : array_like
        List of modes (Noll ordering) to fit.

    normalize : bool, optional
        If True (default), the output is normalized according to [1]. If False,
        the output value ranges [-1, 1] over the mask.

    rho : array_like, optional
        Radial coordinates of the mask array. :attr:`rho` should be 0 at the
        origin and 1 at the edge of the circle.

    theta : array_like, optional
        Angular coordinates of the mask array in radians.

    Returns
    -------
    ndarray
        List of coefficients fit to the supplied OPD over the specified number
        of Zernike modes.

    Example
    -------
    .. code:: pycon

        >>> import numpy as np
        >>> import lentil
        >>> mask = lentil.util.circlemask((256,256),128)
        >>> coeffs = np.random.rand(4)*100e-9
        >>> opd = lentil.zernike.zernike_compose(mask, coeffs)
        >>> fit_coeffs = lentil.zernike.zernike_fit(opd, mask, np.arange(2,4))
        >>> print('Tip/tilt coefficients:', coeffs[1:3])
        >>> print('Fit tip/tilt coefficients:', fit_coeffs)

        Tip/tilt coefficients: [9.69097470e-08 9.94332699e-08]
        Fit tip/tilt coefficients: [9.69545890e-08 9.94781119e-08]

    See Also
    --------
    :func:`zernike_remove` Fit and remove a Zernike basis set from an OPD.

    References
    ----------
    [1] Noll, RJ. Zernike polynomials and atmospheric turbulence. J Opt Soc Am 66, 207-211  (1976).

    """
    opd = np.asarray(opd)
    mask = np.asarray(mask)

    basis = zernike_basis(mask, modes, True, normalize, rho, theta)

    basis = np.linalg.pinv(basis)

    return np.einsum('ij,i->j', basis, opd.ravel())


def zernike_remove(opd, mask, modes, rho=None, theta=None):
    """Fit and remove a Zernike basis set from an OPD.

    Parameters
    ----------
    opd : array_like
        OPD to fit.

    mask : array_like
        Binary mask defining the extent to compute the Zernike basis over.

    modes : array_like
        List of modes (Noll ordering) to remove.

    rho : array_like, optional
        Radial coordinates of the mask array. :attr:`rho` should be 0 at the
        origin and 1 at the edge of the circle.

    theta : array_like, optional
        Angular coordinates of the mask array in radians.

    Returns
    -------
    ndarray
        Residual OPD after the specified Zernike modes have been fit and
        removed.

    See Also
    --------
    :func:`zernike_fit`

    References
    ----------
    [1] Noll, RJ. Zernike polynomials and atmospheric turbulence. J Opt Soc Am 66, 207-211  (1976).

    """
    opd = np.asarray(opd)
    mask = np.asarray(mask)

    coeffs = zernike_fit(opd, mask, modes, rho, theta)
    fit_opd = zernike_compose(mask, coeffs, rho, theta)

    residual = opd - fit_opd

    return residual


def zernike_index(j):
    """Convert a 1-D Noll index to a 2-D radiul and azimuthal index.

    Parameters
    ----------
    j : int
        Noll Zernike index.

    Returns
    -------
    n : int
        Radial Zernike index.

    m : int
        Azimuthal Zernike index.

    """
    if j < 1:
        raise ValueError('Zernike index j must be a positive integer')

    # find the row j is in
    n = int(np.ceil((-1 + np.sqrt(1 + 8*j)) / 2) - 1)
    if n == 0:
        m = 0
    else:
        # find where in the row j is
        k = (n + 1) * (n + 2) / 2
        r = int(j - k - 1)

        if j & 1:  # odd
            sign = -1
        else:
            sign = 1

        if n & 1:  # odd
            row_m = [1, 1]
        else:
            row_m = [0]

        for i in range(int(np.floor(n / 2))):
            row_m.append(row_m[-1] + 2)
            row_m.append(row_m[-1])

        m = row_m[r] * sign

    return m, n


def zernike_coordinates(mask, shift=(0,0), rotate=0):
    """Compute the Zernike coordinate system for a given mask."""

    mask = np.asarray(mask)

    yy, xx = util.mesh(mask.shape, shift)

    r = np.abs(xx+1j*yy)
    rho = r/np.max(r*mask)  # rho is defined to be 1 on the edge of the aperture

    # there is a 90 degree offset since np.angle places 0 degrees up
    angle = (90-rotate) * np.pi/180
    theta = np.angle(xx*np.exp(1j*angle) + 1j*yy*np.exp(1j*angle))

    return rho, theta



import copy
from itertools import combinations
from operator import itemgetter

import numpy as np

from lentil import util
from lentil import fourier
from lentil.plane import Plane, Pupil, Image, Detector, Tilt
from lentil.wavefront import Wavefront

__all__ = ['propagate']


def propagate(planes, wave, weight=None, npix=None, npix_chip=None, oversample=2,
              rebin=True, tilt='phase', interp_phasor=True, flatten=True,
              use_multiprocessing=False):
    """Compute a polychromatic point spread function using Fraunhofer
    diffraction.

    Parameters
    ----------
    planes : list_like
        List of :class:`~lentil.Plane` objects

    wave : array_like
        Array of wavelengths (in meters)

    weight : array_like, optional
        Weight multiple applied to each wavelength slice in :attr:`wave`.
        :attr:`weight` can be relative (for example, when considering the
        optical transmission through the optical system) or absolute (for
        example, when performing radiometrically accurate propagations where
        wavelength-dependent flux at the image plane is known). Must have
        the same length as :attr:`wave`. If not specified, ones are used.

    npix : int or (2,) tuple of ints, optional
        Shape of output plane. If not specified,
        ``npix = OpticalSystem.planes[-1].shape``.

    npix_chip : int or (2,) tuple of ints, optional
        Shape of propagation output plane. If None (default),
        ``npix_chip = npix``. If ``npix_chip != npix``, the propagation
        result is placed in the appropriate location in the output plane.
        npix_chip cannot be larger than npix.

    oversample : int, optional
        Number of times to oversample the output plane. Default is 2.

    rebin : bool, optional
        If ``True``, return the output plane in the sampling given by
        ``pixelscale``, binning down the output plane by a factor of
        ``oversample`` as needed. Note that this operation preserves power
        in the output plane. Default is ``True``.

    tilt : {'phase', 'angle'}, optional
        Propagation tilt handling strategy

        * 'phase' - any tilt present in the Element phase contribution is
          included in the Element*Wavefront product. (Default)
        * 'angle' - any tilt present in the Element phase is removed before
          computing the Element*Wavefront product. The equivalent angular
          tilt is included in the Wavefront's
          :attr:`~lentil.Wavefront.tilt` attribute.

    interp_phasor : bool, optional
        If True (default), the phasor components will be automatically
        interpolated to avoid aliasing and wraparound in the detector plane.
        If False, no checking or interpolation is performed.

    flatten : bool, optional
        If ``True``, the cube of wavelength-dependent output planes is
        flattened into a single 2D array before being returned. If
        ``False``, a cube of output planes is returned. Default is ``True``.

    Returns
    -------
    psf : ndarray
        Resulting point spread function.

    """

    npix = _standardize_shape(npix, default=planes[-1].shape)
    npix_chip = _standardize_shape(npix_chip, default=npix)
    wave = _standardize_bandpass(wave)
    weight = _standardize_bandpass(weight, default=np.ones_like(wave))

    # TODO: this is temporarily commented out for v0.5.0 beta testing
    # if (rebin is True) and (not isinstance(planes[-1], Detector)):
    #     raise AttributeError('rebin is only supported when propagating to a Detector plane')

    _prepare_planes(planes, wave, npix_chip, oversample, tilt, interp_phasor)

    # Create an empty output

    # TODO: temporary image = detector workaround for v0.5.0 beta testing
    output_dtype = np.float64
    #output_dtype = np.float64 if isinstance(planes[-1], Detector) else np.complex128
    output_shape = (npix[0]*oversample, npix[1]*oversample)

    if flatten:
        output = np.zeros(output_shape, dtype=output_dtype)
    else:
        output = np.zeros((len(wave), output_shape[0], output_shape[1]), dtype=output_dtype)

    #_output = np.zeros(output_shape, dtype=np.complex128)

    for n, (wl, wt) in enumerate(zip(wave, weight)):
        if wt > 0:

            w = _propagate_mono(planes, wl, npix_chip, oversample)

            if w.planetype == 'image':

                tiles = []
                for d in range(w.depth):
                    # The shift term is given in terms of (x,y) but we place the chip in
                    # terms of (r,c)
                    shift = np.flip(w.center[d], axis=0)

                    # Compute the chip location
                    data_slice, chip_slice = _chip_insertion_slices(output_shape,
                                                                    (w.data[d].shape[0], w.data[d].shape[1]),
                                                                    shift)
                    if data_slice:
                        tiles.append(imtile(w.data[d], data_slice, chip_slice))

                tiles = consolidate(tiles)

                for tile in tiles:
                    if flatten:
                        output[tile.slice] += np.abs(tile.data)**2 * wt
                    else:
                        output[n][tile.slice] = np.abs(tile.data)**2 * wt



                    # Insert the complex result in the output
                    #if data_slice:
                    #    _output[data_slice] += w.data[d][chip_slice] * np.sqrt(wt)

                #if flatten:
                #    output += np.abs(_output)**2
                #else:
                #    output[n] = np.abs(_output)**2

                #_output[:] = 0

    #output = np.abs(output)**2

            # At this point w.data should always have len == 1
            #if flatten:
            #    output += w.data[0] * wt
            #else:
            #    output[n] = w.data[0] * wt

    if rebin:
        output = util.rebin(output, oversample)

    _cleanup_planes(planes)

    return output


def _prepare_planes(planes, wave, npix, oversample, tilt, interp_phasor):

    # What should happen here as a part of cache_propagate:
    # -----------------------------------------------------
    # * if fit_tilt = True, the tilt in each plane should be fit and removed
    #   from the phase and bookkept in the plane.tilt attribute
    #   - note that this can be done on the un-sliced data for ease of interaction
    #     with ptt_vector
    #   - note further that ptt_vector should be reworked as a property with a
    #     setter that is called at object creation with the mask in case we need
    #     to do some reinterpolation down the line
    #
    # *  Once this is done we can remove ptt_vector from the cache that is
    #    created in cache_propagate() since it won't be needed anymore. As
    #    a matter of fact, we can just turn this into a method that takes
    #    a mask or segmask and returns the vector -> _ptt_vector(mask)
    #
    # * Somewhere in here (probably before we fit and remove tilt), we need
    #   to sort out the sampling and make sure each plane is adequately
    #   sampled. Write up sampling rules for this.
    #
    #

    # note the actual order of operations is probably:
    # 1. sampling
    # 2. set slice
    # 3. fit_tilt (but this can be done over full arrays)
    # TODO: rework this
    # Compute the wavefront shape required to support propagation through the
    # provided list of planes

    # TODO: not sure where we should reinterpolate and how it should be done.

    # The current approach is to find the maximum row and column dimensions of
    # only the Pupil elements in self.planes. This whole approach will have to
    # change once we support more complicated multi-plane propagations.
    #
    # I think what we should really do is find the shape of the first Plane
    # of consequence. What we have hwre works for now though
    shapes = [plane.shape for plane in planes if isinstance(plane, Pupil)]
    npix_wavefront = (max(shapes, key=itemgetter(0))[0],
                      max(shapes, key=itemgetter(1))[1])

    # Loop over the planes to build caches
    for plane in planes:
        # handle the tilt first
        plane.cache['tilt'] = copy.deepcopy(plane.tilt)
        if tilt == 'angle':
            phase, fit_tilt = plane.fit_tilt()
            plane.cache['phase'] = phase
            if fit_tilt:
                plane.cache['tilt'].extend([fit_tilt])
        else:
            plane.cache['phase'] = copy.deepcopy(plane.phase)

        plane.cache['amplitude'] = copy.deepcopy(plane.amplitude)
        plane.cache['pixelscale'] = copy.deepcopy(plane.pixelscale)
        plane.cache['npix_wavefront'] = npix_wavefront
        slc = plane.slice()
        plane.cache['slice'] = slc
        plane.cache['offset'] = plane.slice_offset(slices=slc, shape=plane.shape, indexing='xy')

        # reinterpolate as needed
        # NOTE: if we reinterpolate, we'll have to make sure we provide the correct
        # slices and offsets (using the appropriate reinterpolated mask, probably)
        if interp_phasor:
            # We'll have already computed the interpolation scale when
            # figuring out npix_wavefront so we just need to rescale
            # amplitude and phase here
            pass


def _cleanup_planes(planes):
    for plane in planes:
        plane.cache.clear()


def _propagate_mono(planes, wavelength, npix_chip, oversample):
    """Propagate a monochromatic wavefront from plane to plane through the
    optical system using Fraunhofer diffraction.

    Parameters
    ----------
    w : :class:`~lentil.Wavefront`
        Wavefront object to propagate through the optical system.

    npix : int or (2,) ints
        Shape of output plane.

    oversample : int
       Number of times to oversample the output plane.

    tilt : {'phase', 'angle'}
        * 'phase' - any tilt present in the Element phase contribution is
          included in the Element*Wavefront product.
        * 'angle' - any tilt present in the Element phase is removed
          before computing the Element*Wavefront product. The equivalent
          angular tilt is included in the Wavefront's
          :attr:`~lentil.Wavefront.tilt` attribute.

    Returns
    -------
    field : ndarray
        Resulting complex field propagated though the optical system.

    """

    w = Wavefront(wavelength=wavelength, pixelscale=None,
                  shape=planes[0].cache['npix_wavefront'], planetype=None)

    for plane in planes:

        # Propagate to plane
        if w.planetype is None:
            # free space propagation to first plane
            pass
        elif (w.planetype == 'pupil') and isinstance(plane, Image):
            if plane.pixelscale is not None:
                w = _propagate_pti(w, plane.pixelscale, npix_chip, oversample)
            else:
                pass
        elif (w.planetype == 'image') and isinstance(plane, Pupil):
            pass
        elif (w.planetype == 'pupil') and isinstance(plane, Pupil):
            pass
        elif (w.planetype == 'image') and isinstance(plane, Image):
            pass
        elif isinstance(plane, Plane):
            pass
        else:
            raise TypeError('Unsupported propagation type ', w.planetype, ' to ', plane)

        # Multiply by plane
        w = plane.multiply(w)

    # TODO: More advanced use cases may fail this assertion. We should figure out a
    # way to handle them here
    #assert w.depth == 1

    return w


# TODO:
# 1. accept flatten parameter
# 2. accept some sort of chip processing lambda function
# 3. accept data dtype parameter

def _propagate_pti(w, pixelscale, npix, oversample):

    shift = w.shift(pixelscale, oversample)
    w.tilt = []

    # Integer portion of the shift that will be accounted for
    # later
    fix_shift = np.fix(shift)
    w.center = list(fix_shift)

    # Residual subpixel portion of the shift that is passed to
    # the DFT
    res_shift = shift - fix_shift

    #chip = np.empty((npix[0]*oversample, npix[1]*oversample), dtype=np.complex128)

    alpha = (w.pixelscale * pixelscale) / (w.wavelength * w.focal_length * oversample)

    if shift.shape[0] != w.depth:
        raise ValueError('Dimension mismatch between tilt and wavefront depth')

    for d in range(w.depth):
        # data[d] = fourier.dft2(w.data[d], alpha, npix, res_shift[d])
        w.data[d] = fourier.dft2(w.data[d], alpha, (npix[0]*oversample, npix[1]*oversample),
                                 res_shift[d], offset=w.offset[d], unitary=True)



    return w


def _standardize_shape(shape, default=()):
    if shape is None:
        shape = default
    shape = np.asarray(shape)
    if shape.shape == ():
        shape = np.append(shape, shape)
    return shape


def _standardize_bandpass(vec, default=()):
    if vec is None:
        vec = default
    vec = np.asarray(vec)
    if vec.shape == ():
        vec = vec[np.newaxis, ...]
    return vec


class _iterate_planes:
    def __init__(self, planes):
        self.planes = planes
        self.length = len(planes)
        self.n = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.n < self.length-1:
            plane = self.planes[self.n]
            next_plane = self.planes[self.n+1]
            self.n += 1
            return plane, next_plane
        else:
            raise StopIteration()


class _iterate_planes_reverse:
    def __init__(self, planes):
        self.planes = planes
        self.length = len(planes)
        self.n = 1

    def __iter__(self):
        return self

    def __next__(self):
        if self.n < self.length:
            plane = self.planes[-(self.n+1)]
            next_plane = self.planes[-self.n]
            self.n += 1
            return plane, next_plane
        else:
            raise StopIteration()


def _chip_insertion_slices(npix_canvas, npix_chip, shift):
    npix_canvas = np.asarray(npix_canvas)
    npix_chip = np.asarray(npix_chip)

    # Canvas coordinates of the upper left corner of the shifted chip
    chip_shifted_ul = (npix_canvas / 2) - (npix_chip / 2) + shift

    # Chip slice indices
    chip_top = int(0)
    chip_bottom = int(npix_chip[0])
    chip_left = int(0)
    chip_right = int(npix_chip[1])

    # Canvas insertion slice indices
    canvas_top = int(chip_shifted_ul[0])
    canvas_bottom = int(chip_shifted_ul[0] + npix_chip[0])
    canvas_left = int(chip_shifted_ul[1])
    canvas_right = int(chip_shifted_ul[1] + npix_chip[1])

    # reconcile the chip and canvas insertion indices
    if canvas_top < 0:
        chip_top = -1 * canvas_top
        canvas_top = 0

    if canvas_bottom > npix_canvas[0]:
        chip_bottom -= canvas_bottom - npix_canvas[0]
        canvas_bottom = npix_canvas[0]

    if canvas_left < 0:
        chip_left = -1 * canvas_left
        canvas_left = 0

    if canvas_right > npix_canvas[1]:
        chip_right -= canvas_right - npix_canvas[1]
        canvas_right = npix_canvas[1]

    if np.any(np.array([canvas_bottom, chip_bottom, canvas_right, chip_right]) < 0):
        return None, None
    else:
        return (slice(canvas_top, canvas_bottom), slice(canvas_left, canvas_right)), \
               (slice(chip_top, chip_bottom), slice(chip_left, chip_right))


class imtile:
    def __init__(self, data, global_slice, data_slice):
        self._data = [data]
        self._data_slice = [data_slice]
        self._global_slice = [global_slice]
        self._slice = global_slice

    def join(self, imtile):
        self._data = [*self._data, *imtile._data]
        self._data_slice = [*self._data_slice, *imtile._data_slice]
        self._global_slice = [*self._global_slice, *imtile._global_slice]

        # update the bounding slice
        self._slice = bounding_slice(self._slice, imtile.slice)  # the full output tile slice

    @property
    def slice(self):
        return self._slice

    @property
    def _local_slice(self):
        return [local_slice(self._slice, slc) for slc in self._global_slice]

    @property
    def data(self):
        # this creates the output array and places everything

        data = np.zeros((self.slice[0].stop - self.slice[0].start,
                         self.slice[1].stop - self.slice[1].start),
                        dtype=np.complex128)

        # compute local slices based on slice and global slice

        _local_slice = self._local_slice

        for n in range(len(self._data)):
            data[_local_slice[n]] += self._data[n][self._data_slice[n]]

        return data


def overlap(a, b):
    return a[0].start <= b[0].stop and a[0].stop >= b[0].start and a[1].start <= b[1].stop and a[1].stop >= b[1].start


def bounding_slice(a, b):
    rmin = min(a[0].start, b[0].start)
    rmax = max(a[0].stop, b[0].stop)
    cmin = min(a[1].start, b[1].start)
    cmax = max(a[1].stop, b[1].stop)
    return slice(rmin, rmax), slice(cmin, cmax)


def local_slice(output_slice, chip_slice):
    row_slc = chip_slice[0].start - output_slice[0].start, chip_slice[0].stop - output_slice[0].start
    col_slc = chip_slice[1].start - output_slice[1].start, chip_slice[1].stop - output_slice[1].start
    return slice(*row_slc), slice(*col_slc)


def consolidate(tiles):
    for m, n in combinations(range(len(tiles)), 2):
        if overlap(tiles[m].slice, tiles[n].slice):
            tiles[m].join(tiles[n])
            tiles.pop(n)
            return consolidate(tiles)

    return tiles



import copy

import numpy as np
import scipy

__all__ = ['Spectrum', 'Blackbody', 'Material', 'path_emission', 'path_transmission',
           'planck_radiance', 'planck_exitance', 'vegaflux']

# Constants
H = 6.62606957e-34  # Planck constant, J/s (kg m^2/s^2)
C = 299792456      # Speed of light, m/s
K = 1.3806488e-23  # Boltzmann constant, J/K (kg m^2/K)


class Spectrum:
    """Class for representing spectral quantities.

    :class:`~lentil.radiometry.Spectrum` provides core functionality for
    defining and interacting with spectral data in lentil. Some basic
    arithmetic functions are overloaded, so you can add, multiply, and
    exponentiate :class:`Spectrum` objects

        * with each other
        * by constants
        * with arrays having the same length as :attr:`value`

    Parameters
    ----------
    wave : array_like
        Array of wavelengths.

    value : array_like
        Array of values corresponding to wavelengths in :attr:`wave`.

    waveunit : str, optional
        Wavelength units, as accepted by :func:`Unit`. Default is ``nm``.

    valueunit : str or None, optional
        Value units, as accepted by :func:`Unit`. Default is None.

    Notes
    -----
    When performing arithmetic operations between two
    :class:`Spectrum` objects, the code will behave differently depending on
    the ``wave`` vectors of each :class:`Spectrum`:

        1. If the :attr:`wave` vectors of each :class:`Spectrum` are the same,
        the arithmetic operation is directly performed element-wise on the
        :attr:`value` vectors of each :class:`Spectrum`.

        2. If the :attr:`wave` vectors of each :class:`Spectrum` are
        different, the :attr:`value` vectors of each :class:`Spectrum` are
        interpolated to the same sampling, and then the arithmetic operation is
        performed element-wise on the resulting interpolated :attr:`value`
        vectors. Before interpolating, the :attr:`wave` and :attr:`value`
        vectors are zero-padded to cover the entire wave range represented by
        the union of both :class:`Spectrum` objects. The sampling is determined
        by the smaller sampling of the two :attr:`wave` vectors.

    You can fine tune the sampling, interpolation, and padding methods by
    interfacing directly with the :func:`Spectrum.add` and
    :func:`Spectrum.multiply` methods.

    """

    # TODO: astropy unit cutover
    #   * manage data and units separately
    #   * in __init__, unpack wave and value arrays from unit if passed as a
    #     quantity
    #   * in the wave/value getter, repack as a quantity
    #   * for speed, underlying computations should operate on _wave or _value
    #     and keep the units in mind

    def __init__(self, wave, value, waveunit='nm', valueunit=None):
        self.wave = wave
        self.value = value
        self.waveunit = waveunit
        self.valueunit = valueunit
        self.__array_priority__ = 1.0

        if self._wave.shape != self._value.shape:
            raise ValueError('Wave and value must have the same shape')

    def __add__(self, other):
        return self.add(other)

    def __mul__(self, other):
        return self.multiply(other)

    def __pow__(self, other):
        return self.power(other)

    def __sub__(self, other):
        return self.subtract(other)

    def __truediv__(self, other):
        return self.divide(other)

    __rmul__ = __mul__

    @property
    def wave(self):
        return self._wave

    @wave.setter
    def wave(self, value):
        value = np.asarray(value)

        if np.any(value <= 0):
            raise ValueError('Wavelength values must be greater than zero')

        if not np.alltrue(np.sort(value) == value):
            raise ValueError('Wavelength values must be monotonically increasing')

        if np.any(value[1:] - value[:-1] == 0):
            raise ValueError('Wavelength values must be unique')

        self._wave = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        value = np.asarray(value)
        self._value = value

    @property
    def waveunit(self):
        if self._waveunit is not None:
            return self._waveunit.name
        else:
            return None

    @waveunit.setter
    def waveunit(self, waveunit):
        self._waveunit = Unit(waveunit)

    @property
    def valueunit(self):
        if self._valueunit is not None:
            return self._valueunit.name
        else:
            return None

    @valueunit.setter
    def valueunit(self, valueunit):
        self._valueunit = Unit(valueunit)

    def _ufunc(self, ufunc, other, sampling='min', method='linear', fill_value=0):

        if isinstance(other, (int, float, list, tuple, np.ndarray)):
            wave = self.wave
            try:
                value = ufunc(self.value, other)
            except ValueError:
                raise

        elif isinstance(other, Spectrum):
            wave, self_value, other_value = _interp_common(self, other, sampling,
                                                           method, fill_value)
            value = ufunc(self_value, other_value)
        else:
            raise TypeError(f"can't {ufunc.__name__} Spectrum with object of type "
                            f"{type(other).__name__}")

        return Spectrum(wave, value, self.waveunit, self.valueunit)

    def add(self, other, sampling='min', method='linear', fill_value=0):
        """Add Spectrum and other, element-wise

        Parameters
        ----------
        other : scalar, sequence, or Spectrum
            Any accepted data structure

        sampling : {'min', 'left', 'right', float}, optional
            Method used for computing wavelength sampling when Spectrum
            resampling is required prior to adding

            * 'min' uses the minimum sampling in either Spectrum (Default)
            * 'left' uses the minimum sampling in the left Spectrum
            * 'right' uses the minimum sampling in the right Spectrum
            * <float> uses the float value specified

        method : {'linear', 'quadratic', 'cubic'}, optional
            * 'linear' uses linear interpolation (Default)
            * 'quadratic' uses second-order spline interpolation
            * 'cubic' uses third-order spline interpolation

        fill_value : float or array_like, optional
            * If a float, this value will be used to fill in requested points
              outside of the data range.
            * If a two-element array, the first element is used to fill
              `value_new < value[0]` and the second element is used for
              `value_new > value[-1]`.
            * If not provided, the default is 0.

        Returns
        -------
        result : Spectrum
            Result of the arithmetic operation.

        """
        return self._ufunc(np.add, other, sampling, method, fill_value)

    def multiply(self, other, sampling='min', method='linear', fill_value=0):
        """Multiply Spectrum and other, element-wise

        Parameters
        ----------
        other : scalar, sequence, or Spectrum
            Any accepted data structure

        sampling : {'min', 'left', 'right'} or float, optional
            Method used for computing wavelength sampling when Spectrum
            resampling is required prior to adding

            * 'min' uses the minimum sampling in either Spectrum (Default)
            * 'left' uses the minimum sampling in the left Spectrum
            * 'right' uses the minimum sampling in the right Spectrum
            * <float> uses the float value specified

        method : {'linear', 'quadratic', 'cubic'}, optional
            * 'linear' uses linear interpolation (Default)
            * 'quadratic' uses second-order spline interpolation
            * 'cubic' uses third-order spline interpolation

        fill_value : float or array_like, optional
            * If a float, this value will be used to fill in requested points
              outside of the data range.
            * If a two-element array, the first element is used to fill
              `value_new < value[0]` and the second element is used for
              `value_new > value[-1]`.
            * If not provided, the default is 0.

        Returns
        -------
        result : Spectrum
            Result of the arithmetic operation.

        """
        return self._ufunc(np.multiply, other, sampling, method, fill_value)

    def power(self, other, sampling='min', method='linear', fill_value=0):
        """Spectrum elements raised to powers from other, element-wise

        Parameters
        ----------
        other : scalar, sequence, or Spectrum
            Any accepted data structure

        sampling : {'min', 'left', 'right'} or float, optional
            Method used for computing wavelength sampling when Spectrum
            resampling is required prior to adding

            * 'min' uses the minimum sampling in either Spectrum (Default)
            * 'left' uses the minimum sampling in the left Spectrum
            * 'right' uses the minimum sampling in the right Spectrum
            * <float> uses the float value specified

        method : {'linear', 'quadratic', 'cubic'}, optional
            * 'linear' uses linear interpolation (Default)
            * 'quadratic' uses second-order spline interpolation
            * 'cubic' uses third-order spline interpolation

        fill_value : float or array_like, optional
            * If a float, this value will be used to fill in requested points
              outside of the data range.
            * If a two-element array, the first element is used to fill
              `value_new < value[0]` and the second element is used for
              `value_new > value[-1]`.
            * If not provided, the default is 0.

        Returns
        -------
        result : Spectrum
            Result of the arithmetic operation.

        """
        return self._ufunc(np.power, other, sampling, method, fill_value)

    def subtract(self, other, sampling='min', method='linear', fill_value=0):
        """Subtract Spectrum and other, element-wise

        Parameters
        ----------
        other : scalar, sequence, or Spectrum
            Any accepted data structure

        sampling : {'min', 'left', 'right'} or float, optional
            Method used for computing wavelength sampling when Spectrum
            resampling is required prior to adding

            * 'min' uses the minimum sampling in either Spectrum (Default)
            * 'left' uses the minimum sampling in the left Spectrum
            * 'right' uses the minimum sampling in the right Spectrum
            * <float> uses the float value specified

        method : {'linear', 'quadratic', 'cubic'}, optional
            * 'linear' uses linear interpolation (Default)
            * 'quadratic' uses second-order spline interpolation
            * 'cubic' uses third-order spline interpolation

        fill_value : float or array_like, optional
            * If a float, this value will be used to fill in requested points
              outside of the data range.
            * If a two-element array, the first element is used to fill
              `value_new < value[0]` and the second element is used for
              `value_new > value[-1]`.
            * If not provided, the default is 0.

        Returns
        -------
        result : Spectrum
            Result of the arithmetic operation.

        """
        return self._ufunc(np.subtract, other, sampling, method, fill_value)

    def divide(self, other, sampling='min', method='linear', fill_value=0):
        """Divide Spectrum and other, element-wise

        Parameters
        ----------
        other : scalar, sequence, or Spectrum
            Any accepted data structure

        sampling : {'min', 'left', 'right'} or float, optional
            Method used for computing wavelength sampling when Spectrum
            resampling is required prior to adding

            * 'min' uses the minimum sampling in either Spectrum (Default)
            * 'left' uses the minimum sampling in the left Spectrum
            * 'right' uses the minimum sampling in the right Spectrum
            * <float> uses the float value specified

        method : {'linear', 'quadratic', 'cubic'}, optional
            * 'linear' uses linear interpolation (Default)
            * 'quadratic' uses second-order spline interpolation
            * 'cubic' uses third-order spline interpolation

        fill_value : float or array_like, optional
            * If a float, this value will be used to fill in requested points
              outside of the data range.
            * If a two-element array, the first element is used to fill
              `value_new < value[0]` and the second element is used for
              `value_new > value[-1]`.
            * If not provided, the default is 0.

        Returns
        -------
        result : Spectrum
            Result of the arithmetic operation.

        """
        return self._ufunc(np.divide, other, sampling, method, fill_value)

    def append(self, other, copy=False):
        """Append Spectrum to the end of caller

        Parameters
        ----------
        other : Spectrum
            Spectrum to be appended with

        copy : bool, optional
            If True, the a new :class:`Spectrum` object is returned. If False
            (default), the append operation occurs in place.

        """

        if not isinstance(other, Spectrum):
            raise ValueError('Spectrum objects can only be appended with other '
                             'Spectrum objects')

        if np.any(other.wave <= self.wave):
            raise ValueError()

        if copy:
            new = self.copy()
            new.wave = np.append(new.wave, other.wave)
            new.value = np.append(new.value, other.value)
            return new
        else:
            self.wave = np.append(self.wave, other.wave)
            self.value = np.append(self.value, other.value)

    def copy(self):
        return copy.deepcopy(self)

    def sample(self, wave, method='linear', fill_value=0, waveunit='nm'):
        """Sample the :class:`Spectrum` at a set of desired wavelengths.

        Parameters
        ----------
        wave : array_like or float
            Wavelength set for sampling.

        method : {'linear', 'quadratic', 'cubic'}, optional
            * 'linear' uses linear interpolation (default)
            * 'quadratic' uses second order spline interpolation
            * 'cubic' uses third order spline interpolation

        fill_value : float or array_like, optional
            * If a float, this value will be used to fill in requested
              points outside of the data range.
            * If a two-element array, the first element is used to fill
              `value_new < value[0]` and the second element is used for
              `value_new > value[-1]`.
            * If not provided, the default is 0.

        waveunit : str
            Wavelength units, as accepted by :func:`Unit`. Default is ``nm``.

        Returns
        -------
        ndarray
            An array of sampled values.

        """

        if waveunit != self.waveunit:
            self.to(waveunit)

        interp = scipy.interpolate.interp1d(self.wave, self.value, kind=method,
                                            copy=False, bounds_error=False,
                                            fill_value=fill_value)

        return interp(wave)

    def resample(self, wave, method='linear', fill_value=0, waveunit='nm'):
        """Sample the :class:`Spectrum` object at a set of desired wavelengths
        and overwrite the object's :attr:`wave` and :attr:`value` attributes.
        This method can be viewed as a destructive version of
        :func:`Spectrum.sample`

        Parameters
        ----------
        wave : array_like or float
            Wavelength set for sampling.

        method : {'linear', 'quadratic', 'cubic'}, optional
            * 'linear' uses linear interpolation (default)
            * 'quadratic' uses second order spline interpolation
            * 'cubic' uses third order spline interpolation

        fill_value : float or array_like, optional
            * If a float, this value will be used to fill in requested points
              outside of the data range.
            * If a two-element array, the first element is used to fill
              `value_new < value[0]` and the second element is used for
              `value_new > value[-1]`.
            * If not provided, the default is 0.

            Wavelength units, as accepted by :func:`Unit`. Default is ``nm``.

        """
        self.value = self.sample(wave, method=method, fill_value=fill_value,
                                 waveunit=waveunit)
        self.wave = wave
        self.waveunit = waveunit

    def bin(self, wave, interp_method='simps', ends='symmetric', preserve_power=True,
            sample_method='linear', fill_value=0, waveunit='nm'):
        r"""Compute a binned representation of the :class:`Spectrum` data at a
        given array of central wavelengths.

        The general algorithm computes bin edges, sampling the :class:`Spectrum`
        values at these edges, and numerically integrating within the edges of
        each bin. For :math:`n` bins (centered at the wavelength values in the
        ``wave`` array), there will be :math:`n+1` edges. The edges at each bin
        centered at :math:`\lambda_k` are given by

        .. math::

            e_{\mbox{left}} = \lambda_k - \frac{\lambda_k - \lambda_{k-1}}{2}

            e_{\mbox{right}} = \lambda_k + \frac{\lambda_{k+1} - \lambda_k}{2}

        Special treatment is required for the first and last edges
        (:math:`\lambda_k =` ``wave[0]`` and :math:`\lambda_k =` ``wave[-1]``)
        since :math:`\lambda_{k-1}` and :math:`\lambda_{k+1}` are undefined,
        respectively.

        Two different first/last bin edge methods are available:

        .. image:: /_static/img/api/radiometry/spectrum_bin.png

        ========= =============================================== ==================================================================================
        ends      Left edge for ``wave[0]``                       Right edge for ``wave[-1]``
        ========= =============================================== ==================================================================================
        symmetric :math:`\lambda_1-\frac{\lambda_2-\lambda_1}{2}` :math:`\lambda_{\mbox{end}}+\frac{\lambda_{\mbox{end}}-\lambda_{\mbox{end}-1}}{2}`
        inside    :math:`\lambda_1`                               :math:`\lambda_{\mbox{end}}`
        ========= =============================================== ==================================================================================

        Parameters
        ----------
        wave : array_like
            Central wavelength set for binning. At least two wavelengths must be
            provided.

        interp_method : {'simps', 'trapz'}, optional
            Numerical integration method. 'simps' (default) is better for smooth
            data while 'trapz' is better for linear data.

        ends : {'symmetric', 'inside'}, optional
            Method for handling the first and last bin edge value. Default is
            'symmetric'. See the notes below for details on the available
            methods.

        preserve_power : bool, optional

        sample_method : {'linear', 'quadratic', 'cubic'}, optional
            * 'linear' uses linear interpolation (default)
            * 'quadratic' uses second order spline interpolation
            * 'cubic' uses third order spline interpolation

        fill_value : float or array_like, optional
            * If a float, this value will be used to fill in requested points
              outside of the data range.
            * If a two-element array, the first element is used to fill
              `value_new < value[0]` and the second element is used for
              `value_new > value[-1]`.
            * If not provided, the default is 0.

        Returns
        -------
        value : ndarray
            The binned values, same shape as wave.

        """

        wave = np.asarray(wave)

        if wave.size < 2:
            raise ValueError('Spectrum.bin requires a minimum of two wavelengths.\n'
                             'If this Spectrum must be represented by a single '
                             'wavelength, consider using Spectrum.integrate() instead.')

        if waveunit != self.waveunit:
            self.to(waveunit)

        if interp_method == 'trapz':
            dx = np.diff(wave)/2
            x = wave[0:-1] + dx

            if ends == 'symmetric':
                x = np.concatenate([[wave[0]-dx[0]], x, [wave[-1]+dx[-1]]])
            elif ends == 'inside':
                x = np.concatenate([[wave[0]], x, [wave[-1]]])
            else:
                raise ValueError('Unknown ends ', ends)

            # sample
            f = self.sample(x, method=sample_method, fill_value=fill_value)

            # apply the chained trapezoidal rule
            bins = np.array([])
            for k in range(1, f.size):
                bins = np.append(bins, 0.5*(f[k-1]+f[k])*(x[k]-x[k-1]))

        elif interp_method == 'simps':
            dx = np.diff(wave)/2
            wave_mid = wave[0:-1] + dx

            # interleave the wave and wave_mid arrays
            x = np.empty((wave.size + wave_mid.size,), dtype=wave.dtype)
            x[0::2] = wave
            x[1::2] = wave_mid

            if ends == 'symmetric':
                x = np.concatenate([[wave[0]-dx[0]], x, [wave[-1]+dx[-1]]])
            elif ends == 'inside':
                x = np.insert(x, 1, x[0] + (x[1]-x[0])/2)
                x = np.insert(x, -1, x[-1] + (x[-2]-x[-1])/2)
            else:
                raise ValueError('Unknown ends ', ends)

            # sample
            f = self.sample(x, method=sample_method, fill_value=fill_value)

            # apply the chained simpson's rule
            bins = np.array([])
            for k in range(1, x.size, 2):
                bins = np.append(bins, ((x[k+1]-x[k-1])/6) * (f[k-1]+4*f[k]+f[k+1]))

        else:
            raise ValueError('Unknown method ', interp_method)

        if preserve_power:
            norm_factor = self.integrate(np.min(wave), np.max(wave), method=interp_method)/np.sum(bins)
            bins *= norm_factor

        return bins

    def integrate(self, start=None, end=None, method='simps'):
        """Compute the integrated value between ``start`` and ``end``.

        Parameters
        ----------
        start : float, optional
            Lower wavelength bound. If not specified (default),
            ``min(self.wave)`` is used.

        end : float, optional
            Upper wavelength bound. If not specified (default),
            ``max(self.wave)`` is used.

        method : {'simps', 'trapz'}, optional
            Numerical integration method. 'simps' (default) is better for smooth
            data while 'trapz' is better for linear data.

        Returns
        -------
        result : float
            Integral as approximated by ``interp_method``

        """
        if start is None:
            start = np.min(self.wave)
        if end is None:
            end = np.max(self.wave)

        indices = np.intersect1d(np.where(self.wave >= start),
                                 np.where(self.wave <= end))
        wave = self.wave[indices]
        value = self.value[indices]

        if method == 'simps':
            result = scipy.integrate.simps(value, wave)
        elif method == 'trapz':
            result = np.trapz(value, wave)
        else:
            raise ValueError('Unknown method ', method)

        return result

    def ends(self, tol=1e-4):
        """Locate the indices defining the continuous nonzero portion of the
        :attr:`~radiometry.Spectrum.value`

        The first and last indices where the normalized :attr:`Spectrum.value`
        is greater than :attr:`tol` define the bounds of the retained portion of
        the :class:`Spectrum`.

        Parameters
        ----------
        tol : float, optional
            Relative tolerance used to find ends. Default is 1e-4

        """
        if max(self.value) <= 0:
            raise ValueError('Spectrum.trim is only available when '
                             'max(Spectrum.value) > 0')
        normval = self.value/max(self.value)
        index = np.where(normval > tol)
        index_min = index[0][0]
        index_max = index[0][-1]
        return index_min, index_max

    def trim(self, tol=1e-4):
        """Trim the zero or near-zero ends off the :class:`Spectrum` object.

        The first and last indices where the normalized :attr:`Spectrum.value`
        is greater than :attr:`tol` define the bounds of the retained portion of
        the :class:`Spectrum`.

        Parameters
        ----------
        tol : float, optional
            Relative tolerance used to find ends. Default is 1e-4

        Note
        ----
        If :attr:`~radiometry.Spectrum.value` is all zeros, no trim operation is
        performed and the Spectrum remains unchanged.

        """
        # If the value array is all zeros, do nothing!
        if not self.value.any():
            return

        index_min, index_max = self.ends(tol)
        self.wave = self.wave[index_min:index_max+1]
        self.value = self.value[index_min:index_max+1]

    def crop(self, min_wave, max_wave):
        """Crop a :class:`Spectrum` object."""

        if min_wave > self.wave[0]:
            indx = np.where(min_wave > self.wave)
            self.wave = np.delete(self.wave, indx)
            self.value = np.delete(self.value, indx)

        if max_wave < self.wave[-1]:
            indx = np.where(max_wave < self.wave)
            self.wave = np.delete(self.wave, indx)
            self.value = np.delete(self.value, indx)

    def pad(self, ends, sampling='min', mode='constant', **kwargs):
        """Pad a Spectrum

        Parameters
        ----------
        ends : sequence or array_like, optional
            Wavelength ends to pad to. If None (default), a single value is
            padded to each end of the Spectrum according to the sampling and
            mode parameters.

        sampling : 'min' or float, optional
            * 'min' (Default)
            * float

        mode : {'constant', 'edge'}, optional
            * 'constant' pads with a constant value (Default)
            * 'edge' pads with the edge values

        values : sequence or int, optional
            Used in 'constant'. The (before, after) values to set the padded
            values for each end. int is a shortcut for before = after. Default
            is 0.

        """

        if mode == 'constant':
            if 'values' in kwargs:
                values = np.asarray(kwargs['values'])
                if values.shape == ():
                    values = np.append(values, values)
            else:
                values = (0, 0)
        elif mode == 'edge':
            values = np.array([self.value[0], self.value[-1]])
        else:
            raise ValueError('Unknown mode', mode)

        dwave = _sampling(self.wave, sampling)

        minwave, maxwave = self.wave.min(), self.wave.max()

        ends = np.asarray(ends)

        nleft = int(np.ceil((minwave - ends[0])/dwave)) + 1
        nright = int(np.ceil((ends[1]-maxwave)/dwave)) + 1

        leftwave = np.linspace(ends[0], minwave, nleft)
        leftwave = np.delete(leftwave, -1)
        leftvalue = values[0] * np.ones(leftwave.shape)

        rightwave = np.linspace(maxwave, ends[1], nright)
        rightwave = np.delete(rightwave, 0)
        rightvalue = values[1] * np.ones(rightwave.shape)

        self.wave = np.hstack((leftwave, self.wave, rightwave))
        self.value = np.hstack((leftvalue, self.value, rightvalue))

    def asarray(self):
        """Return :attr:`wave` and :attr:`value` as an array.

        Returns
        -------
        array : ndarray
            Spectral data as an array with ``array[0] = wave`` and ``array[1] = value``.
        """
        return np.array((self.wave, self.value))

    def to(self, *args):
        """Set new wavelength and/or value units. In addition to supporting
        single ``wave`` or ``value`` unit conversion, this method accepts two
        unit arguments to perform simultaneous ``wave`` and ``value`` unit
        conversions.

        Parameters
        ----------
        unit : str
            Unit to convert to, as accepted by :func:`Unit`.

        """
        for unit in args:

            if unit.lower() in ['m', 'um', 'nm', 'angstrom']:
                if self.valueunit in ['photlam', 'flam', 'wlam']:
                    # if the current valueunit are relative to waveunit, we
                    # need to convert to the requested waveunit and update the
                    # valueunit appropriately
                    self.wave = self.wave * self._waveunit.to(unit)
                    self.value = self.value / self._waveunit.to(unit)
                elif self.valueunit is None:
                    # the current valueunit is None, so we assume it is
                    # independent of waveunit and only convert to the
                    # requested waveunit
                    self.wave = self.wave * self._waveunit.to(unit)
                # Update waveunit
                self.waveunit = unit

            elif unit.lower() in ['photlam', 'flam', 'wlam']:
                if self.valueunit is None:
                    raise TypeError("Can't convert from None valueunit to " + unit)
                else:
                    # All conversions are done with wavelength in terms of meters
                    wave = self.wave * self._waveunit.to('meter')
                    value = self.value / self._waveunit.to('meter')

                    # Do the value conversion
                    self.value = self._valueunit.to(value, unit, wave) / Meter().to(self.waveunit)

                    # Update fluxunit
                    self.valueunit = unit

            else:
                raise ValueError('Unknown unit')

    @classmethod
    def from_csv(cls, filename, waveunit, valueunit, zero_negatives=True,
                 header_rows=1):
        """Create a :class:`Spectrum` from a csv file.

        Parameters
        ----------
        filename : str
            csv file to load

        waveunit : str
            Wavelength units, as accepted by :func:`Unit`.

        valueunit : str
            Value units, as accepted by :func:`Unit`.

        zero_negatives : bool, optional
            If true (default), negative values are replaced with zeros.

        header_rows : int, optional
            Number of header rows in csv file to ignore. Default is 1.
        """

        dat = np.genfromtxt(filename, delimiter=',', skip_header=header_rows)
        if zero_negatives:
            dat[dat[:, 1] < 0, 1] = 0

        return cls(dat[:, 0], dat[:, 1], waveunit, valueunit)


def _sampling(wave, method='min'):
    """Compute wave sampling using to specified method.

    Parameters
    ----------
    wave : array_like
        Single wavelength array or 2D collection of wavelength arrays

    method: {'min', 'left', 'right'} or float
        Method used for computing wavelength sampling when Spectrum resampling
        is required prior to adding

        * 'min' uses the minimum sampling in wave
        * 'left' uses the minimum sampling in the left Spectrum
        * 'right' uses the minimum sampling in the right Spectrum
        * <float> uses the float value specified
    """

    if method == 'min':

        if isinstance(wave, (list, tuple)):
            dwave = np.array([])
            for w in wave:
                dwave = np.append(dwave, np.diff(w).min())

            return dwave.min()

        elif isinstance(wave, np.ndarray):
            return np.diff(wave).min()

    elif method == 'left':
        if len(wave) != 2:
            raise ValueError()
        return _sampling(wave[0], method='min')

    elif method == 'right':
        if len(wave) != 2:
            raise ValueError()
        return _sampling(wave[1], method='min')

    elif np.isscalar(method):
        return method

    else:
        raise ValueError('Unknown sampling method', method)


def _intersect(subset, superset):
    """Return the superset indices where the subset overlaps based on its
    subset.min() and subset.max().

    Both subset and superset are assumed to be monotonically increasing

    """
    return np.where((superset >= subset.min()) & (superset <= subset.max()))


def _interp_common(s1, s2, sampling, method, fill_value):
    """Return a common wavelength basis with appropriately interpolated and
    padded value arrays for s1 and s2.

    Parameters
    ----------
    s1, s2 : Spectrum objects

    sampling : {'min', 'left', 'right'} or float
        Method used for computing wavelength sampling when Spectrum resampling
        is required prior to adding

        * 'min' uses the minimum sampling in either Spectrum
        * 'left' uses the minimum sampling in the left Spectrum
        * 'right' uses the minimum sampling in the right Spectrum
        * <float> uses the float value specified

    method : {'linear', 'quadratic', 'cubic'}
        Interpolation method used for sampling Spectrum values.

        * 'linear' uses linear interpolation
        * 'quadratic' uses second-order spline interpolation
        * 'cubic' uses third-order spline interpolation

    fill_value : float
        Value used to fill in requested points outside of the data range.

    Returns
    -------
    commonwave : ndarray

    s1_value : ndarray

    s2_value : ndarray

    """

    # compute a common wavelength array that spans both spectrum and has the
    # desired sampling
    minwave = min(s1.wave.min(), s2.wave.min())
    maxwave = max(s1.wave.max(), s2.wave.max())

    dwave = _sampling((s1.wave, s2.wave), sampling)

    num = int(np.ceil((maxwave - minwave)/dwave))
    commonwave = np.linspace(minwave, maxwave, num + 1)

    # get the portion of commonwave that corresponds to the two spectrum objects
    s1_index = _intersect(s1.wave, commonwave)
    s1_wave = commonwave[s1_index]

    s2_index = _intersect(s2.wave, commonwave)
    s2_wave = commonwave[s2_index]

    # sample each Spectrum at the requested sampling
    s1_samplevalue = s1.sample(s1_wave, method=method, fill_value=fill_value)
    s2_samplevalue = s2.sample(s2_wave, method=method, fill_value=fill_value)

    # create nominal value arrays
    s1_value = fill_value * np.ones(commonwave.shape)
    s2_value = fill_value * np.ones(commonwave.shape)

    # insert the sampled values into the appropriate slots
    s1_value[s1_index] = s1_samplevalue
    s2_value[s2_index] = s2_samplevalue

    return commonwave, s1_value, s2_value


class Blackbody(Spectrum):
    """Class for representing a blackbody emitter.

    Parameters
    ----------
    wave : array_like
        Array of wavelengths or wavenumbers

    temp : float
        Temperature in K

    waveunit : str
        Wavelength units, as accepted by :func:`Unit`. Default is ``nm``.

    valueunit : str
        Flux units, as accepted by :func:`Unit`. Default is ``photlam``.

    Attributes
    ----------
    value : ndarray
       Spectral radiance computed with :func:`planck_radiance`

    Examples
    --------
    Create a Blackbody object with wavelength range 400-4000 nm and a temperature of
    4000K:

    .. code:: pycon

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> import lentil as mo
        >>> wave = np.arange(400,4000)
        >>> temp = 5000
        >>> src = mo.radiometry.Blackbody(wave,temp,waveunit='nm')
        >>> plt.plot(src.wave, src.value), plt.grid()

    .. image:: /_static/img/api/radiometry/blackbody.png
        :width: 400px

    """

    def __init__(self, wave, temp, waveunit='nm', valueunit='photlam'):
        wave = np.asarray(wave)
        flux = planck_radiance(wave, temp, waveunit, valueunit)
        self.temp = temp
        self.sample_fn = planck_radiance

        super().__init__(wave, flux, waveunit, valueunit)

    def sample(self, wave, waveunit='nm', *args, **kwargs):
        # sample_fn will be planck_radiance if the Blackbody was created from its
        # standard constructor. If instead, Blackbody was created from the vegamag
        # classmethod, sample_fn will be planck_exitance
        return self.sample_fn(wave, self.temp, waveunit, valueunit=self.valueunit)

    def sample_vegamag(self, wave, temp, waveunit, valueunit):
        # Get Vega zero point data for requested band
        E0, wave0 = vegaflux(self.band, waveunit)

        # Compute exitance of Vega at equivalent wavelength in the observing
        # band
        M0 = planck_exitance(wave0, temp, waveunit, valueunit)

        # Compute the source exitance over the desired wavelengths
        M = planck_exitance(wave, temp, waveunit, valueunit)

        # Scale the source irradiance by the requested magnitude
        E = E0 * (M/M0)*10**(-0.4*self.mag)

        return E

    # TODO: this should really return a different type. when that is done, remove
    # the sample_fn garbage and move sample_vegamag to the new object's sample
    # method (should actually create a flux method that takes band, mag, temp,
    # etc and call that in both the constructor and the sample method)
    @classmethod
    def vegamag(cls, wave, temp, mag, band, waveunit='nm', valueunit='photlam'):
        r""" Create a :class:`Blackbody` with its irradiance scaled to an
        apparent magnitude.

        Parameters
        ----------
        wave : array_like
            Array of wavelengths or wavenumbers

        temp : float
            Temperature in K

        mag : float
            Apparent magnitude

        band : str
            Observing band closest to the wavelengths in :attr:`wave`. See
            :func:`vegaflux` for supported bands.

        waveunit : str
            Wavelength units, as accepted by :func:`Unit`. Default is ``nm``.

        valueunit : str
            Flux units, as accepted by :func:`Unit`. Default is ``photlam``.

        Examples
        --------
        Create a Blackbody object with wavelength range 400-900 nm, a temperature of
        4000K, and a visible magnitude of 2:
        
        .. code:: pycon

            >>> import numpy as np
            >>> import matplotlib.pyplot as plt
            >>> import lentil as mo
            >>> wave = np.arange(400,900)
            >>> temp = 5000
            >>> mag = 2
            >>> band = 'V'  # Chosen for best overlap with wavelength range
            >>> src = mo.radiometry.Blackbody.vegamag(wave,temp,mag,band,waveunit='nm')
            >>> plt.plot(src.wave, src.value), plt.grid()

        .. image:: /_static/img/api/radiometry/blackbody_vegamag.png
            :width: 400px

        Notes
        -----
        This class implements a special formulation of a :class:`Blackbody`
        which computes irradiance for a star given a magnitude :math:`mag` in
        an observing band chosen from :func:`vegaflux`. The general calculation
        performed here is as follows:

        1. Look up zero point irradiance :math:`E_{\mbox{Vega}}` and central
           wavelength :math:`\lambda_{0,\mbox{Vega}}` in the desired observing
           band
        2. Compute exitance :math:`M_0` for the desired star at
           :math:`\lambda_{0,\mbox{Vega}}`
        3. Compute exitance :math:`M` for the desired star over the requested
           wavelength range
        4. Compute apparent magnitude scaling factor
           :math:`f = 10^{-0.4\mbox{mag}}`
        5. Compute star irradiance as
           :math:`E = E_{\mbox{Vega}} \ f \left(\frac{M}{M_0}\right)`

        """

        wave = np.asarray(wave)

        # Get Vega zero point data for requested band
        E0, wave0 = vegaflux(band, waveunit)

        # Compute exitance of Vega at equivalent wavelength in the observing
        # band
        M0 = planck_exitance(wave0, temp, waveunit, valueunit='photlam')

        # Compute the source exitance over the desired wavelengths
        M = planck_exitance(wave, temp, waveunit, valueunit='photlam')

        # Scale the source irradiance by the requested magnitude
        E = E0 * (M/M0)*10**(-0.4*mag)

        # Construct a blackbody object
        self = cls(wave, temp, waveunit, valueunit)
        self.value = E

        # Populate this object with additional attributes
        self.mag = mag
        self.band = band
        self.sample_fn = self.sample_vegamag

        return self


class Material:
    """Class for representing a material with radiometric properties.

    Parameters
    ----------
    transmission : :class:`~lentil.radiometry.Spectrum` or float, optional
        Spectral transmission. If not specified, a default transmission is
        created which has no radiometric effect.

    emission : :class:`~lentil.radiometry.Spectrum` or float, optional
        Spectral thermal emission of the element. If not specified, a default
        emission is created which has no radiometric contribution.

    contam : float or :class:`~lentil.radiometry.Spectrum`, optional
        Material contamination factor. Can be used for representing EOL/BOL
        optical properties. Should be within [0,1]. Both
        :attr:`~lentil.radiometry.Material.transmission` and
        :attr:`~lentil.radiometry.Material.emission` are multiplied by contam
        before being returned.

    Note
    ----
    There is no separate reflection attribute. Reflective optics should use the
    :attr:`~lentil.radiometry.Material.transmission` attribute to represent
    reflectivity.

    See Also
    --------
    * :func:`lentil.radiometry.path_transmission`
    * :func:`lentil.radiometry.path_emission`

    """
    def __init__(self, transmission=1, emission=0, contam=1):
        self.transmission = transmission
        self.emission = emission
        self.contam = contam

    @property
    def transmission(self):
        return self.contam * self._transmission

    @transmission.setter
    def transmission(self, value):
        self._transmission = value

    @property
    def emission(self):
        # TODO: make sure this is correct
        return self.contam * self._emission

    @emission.setter
    def emission(self, value):
        self._emission = value


def path_transmission(iterable):
    """Construct a common transmission from an iterable of :class:`Spectrum` or
    :class:`Material` objects or numeric types

    Parameters
    ----------
    iterable: list_like
        List of :class:`Spectrum` or :class:`Material` objects or numeric types

    Returns
    -------
    transmission: :class:`~lentil.radiometry.Spectrum`

    """
    transmission = 1
    for item in iterable:
        if isinstance(item, Material):
            transmission = transmission * item.transmission
        else:
            transmission = transmission * item
    return transmission


def path_emission(iterable, emission=0):
    """Construct a common emission from an iterable of :class:`Material` objects

    Parameters
    ----------
    iterable: list_like
        List of :class:`Material` objects

    emission: scalar or :class:`~lentil.radiometry.Spectrum`, optional
        Emission collected upstream of the iterable path. This could be used to
        represent things like zodiacal background or warm cavities emitting
        into the first element of a path.

    Returns
    -------
    emission: :class:`~lentil.radiometry.Spectrum`

    """
    for item in iterable:
        emission = (emission * item.transmission) + item.emission
    return emission


def planck_exitance(wave, temp, waveunit='nm', valueunit='wlam'):
    r"""Compute the Planck law spectral exitance from a blackbody radiator at
    the given temperature.

    .. math::

        M_{\lambda}(T) = \frac{2\pi hc^2}{\lambda^5 \exp \left(\frac{hc}{\lambda k T}\right)-1}

    Parameters
    ----------
    wave : array_like or float
        Wavelength or array of wavelengths

    temp : float
        Blackbody temperature in K

    waveunit : str
        Wavelength units, as accepted by :func:`Unit`.

    valueunit : str
        Flux units, as accepted by :func:`Unit`.

    Returns
    -------
    ndarray
        Spectral exitance in ``valueunit``.

    """

    # convert wave to meters
    wave = wave * Unit(waveunit).to('meter')

    # compute flux in W m^-2 sr^-1 m^-1
    flux = 2*np.pi*H*C**2/(wave**5*(np.exp(H*C/(wave*K*temp))-1))

    # do flux conversion (if necessary)
    if valueunit == 'wlam':
        # just convert flux back to W m^-2 sr^-1 <waveunit>^-1
        return flux / Meter().to(waveunit)
    else:
        # convert to desired fluxunit and back to waveunit
        return Wlam().to(flux, valueunit, wave) / Meter().to(waveunit)


def planck_radiance(wave, temp, waveunit='nm', valueunit='wlam'):
    r"""Compute the Planck law spectral radiance from a blackbody radiator at
    the given temperature.

    .. math::

        M_{\lambda}(T) = \frac{2hc^2}{\lambda^5 \exp \left(\frac{hc}{\lambda k T}\right)-1}

    Parameters
    ----------
    wave : array_like or float
        Wavelength or array of wavelengths

    temp : float
        Blackbody temperature in K

    waveunit : str
        Wavelength units, as accepted by :func:`Unit`. Default is ``nm``.

    valueunit : str
        Flux units, as accepted by :func:`Unit`. Default is ``wlam``.

    Returns
    -------
    ndarray
        Spectral radiance in ``valueunit sr^-1``.

    """
    # convert wave to meters
    wave = wave * Unit(waveunit).to('meter')

    # compute flux in W m^-2 m^-1
    flux = 2*H*C**2/(wave**5*(np.exp(H*C/(wave*K*temp))-1))

    # do flux conversion (if necessary)
    if valueunit == 'wlam':
        # just convert flux back to W m^-2 <waveunit>^-1
        return flux / Meter().to(waveunit)
    else:
        # convert to desired valueunit and back to waveunit
        return Wlam().to(flux, valueunit, wave) / Meter().to(waveunit)


def vegaflux(band, waveunit='nm', valueunit='photlam'):
    """Get the flux of Vega in a desired observing band.

    ====== ================== =======
    Band   Central Wavelength Flux
    ====== ================== =======
    ``U``  360 nm             1790 Jy
    ``B``  438 nm             4036 Jy
    ``V``  545 nm             3636 Jy
    ``R``  641 nm             3064 Jy
    ``I``  798 nm             2416 Jy
    ``J``  1220 nm            1589 Jy
    ``H``  1630 nm            1021 Jy
    ``K``  2190 nm            640 Jy
    ``W1`` 3353 nm            310 Jy
    ``W2`` 4603 nm            172 Jy
    ``W3`` 11561 nm           31.7 Jy
    ``W4`` 22088 nm           8.36 Jy
    ====== ================== =======

    Parameters
    ----------
    band : str
        Observing band.

    waveunit : str
        Wavelength units, as accepted by :func:`Unit`. Default is ``nm``.

    valueunit : str
        Flux units, as accepted by :func:`Unit`. Default is ``photlam``.

    Returns
    -------
    flux : float
        Vega flux in :attr:`band` expressed in :attr:`valueunit`

    wave : float
        Central wavelength in :attr:`band` expressed in :attr:`waveunit`

    References
    ----------
    [1] Bessell et al. (1998) Johnson-Cousins-Class System

    [2] Wright et al. (2010) Wide-Field Infrared Survey Explorer

    """
    vega = {
        # band    nm                   Jy
        'U':  {'wave': 360e-9,   'flux': 1790},
        'B':  {'wave': 438e-9,   'flux': 4036},
        'V':  {'wave': 545e-9,   'flux': 3636},
        'R':  {'wave': 641e-9,   'flux': 3064},
        'I':  {'wave': 798e-9,   'flux': 2416},
        'J':  {'wave': 1220e-9,  'flux': 1589},
        'H':  {'wave': 1630e-9,  'flux': 1021},
        'K':  {'wave': 2190e-9,  'flux': 640},
        'W1': {'wave': 3353e-9,  'flux': 310},
        'W2': {'wave': 4603e-9,  'flux': 172},
        'W3': {'wave': 11561e-9, 'flux': 31.7},
        'W4': {'wave': 22088e-9, 'flux': 8.36}
        }

    band = band.upper()

    if band not in vega:
        raise ValueError('Unknown band: ' + band)

    wave = vega[band]['wave']
    flux = vega[band]['flux']

    flux = flux * 1e-26                 # Jy         -> W/m^2 Hz
    flux = flux * C/(wave**2)           # W/m^2 Hz   -> W/m^2 m
    flux = flux * wave/(H*C)            # W/m^2 m    -> ph/s m^2 m

    # do flux conversion (if necessary)
    if valueunit == 'photlam':
        # just convert ph/s m^2 m^-1 -> ph/sm^-2 <waveunit>^-1
        flux = flux / Meter().to(waveunit)
    else:
        # convert to desired valueunit and back to waveunit. Note that wave
        # is still in terms of meters here.
        flux = Photlam().to(flux, valueunit, wave) / Meter().to(waveunit)

    wave = wave * Meter().to(waveunit)  # m -> <waveunit>

    return flux, wave


def Unit(name):
    r"""Generate a Unit object.

    **Wavelength Units**

    ===================== ================== ==========================
    Name                  Class Object       Units
    ===================== ================== ==========================
    ``m``, ``meter``      :class:`Meter`     SI base unit
    ``um``, ``micron``    :class:`Micron`    :math:`10^{-6}\ \mbox{m}`
    ``nm``, ``nanometer`` :class:`Nanometer` :math:`10^{-9}\ \mbox{m}`
    ``angstrom``          :class:`Angstrom`  :math:`10^{-10}\ \mbox{m}`
    ===================== ================== ==========================

    **Flux Units**

    =========== ================= ===========================================
    Name        Class Object      Units
    =========== ================= ===========================================
    ``photlam`` :class:`Photlam`  :math:`\mbox{photons s}^{-1} \mbox{m}^{-2}`
    ``wlam``    :class:`Wlam`     :math:`\mbox{W m}^{-2}`
    ``flam``    :class:`Flam`     :math:`\mbox{erg s}^{-1} \mbox{cm}^{-2}`
    =========== ================= ===========================================

    Parameters
    ----------
    name : str
        Wavelength or flux units as accpeted by :func:`Unit`.

    Returns
    -------
    Unit or ``None``
        Unit object. ``None`` means unitless.

    """
    if name is not None:
        if name.lower() in ['meter', 'm']:
            return Meter()
        elif name.lower() in ['um', 'micron']:
            return Micron()
        elif name.lower() in ['nm', 'nanometer']:
            return Nanometer()
        elif name.lower() == 'angstrom':
            return Angstrom()
        elif name.lower() == 'photlam':
            return Photlam()
        elif name.lower() == 'flam':
            return Flam()
        elif name.lower() == 'wlam':
            return Wlam()
        else:
            raise ValueError('Invalid unit: ' + name)
    else:
        return None


class _Unit:
    def __str__(self):
        return self.name


class Meter(_Unit):
    """Meter waveunit."""
    name = 'm'

    @staticmethod
    def to(waveunit):
        """Convert waveunit.

        ===== ============ ======================
        From  To           Conversion
        ===== ============ ======================
        ``m`` ``m``        :math:`1`
        ``m`` ``um``       :math:`10^{-6}` ``m``
        ``m`` ``nm``       :math:`10^{-9}` ``m``
        ``m`` ``angstrom`` :math:`10^{-10}` ``m``
        ===== ============ ======================

        Parameters
        ----------
        waveunit : str
            Wavelength units, as accpeted by :func:`Unit`.

        Returns
        -------
        int
            :attr:`wave` scaling factor

        """
        if waveunit.lower() in ['m', 'meter']:
            return 1
        elif waveunit.lower() in ['um', 'micron']:
            return 1e6
        elif waveunit.lower() in ['nm', 'nanometer']:
            return 1e9
        elif waveunit.lower() == 'angstrom':
            return 1e10
        else:
            raise ValueError('Unknown waveunit: ' + waveunit)


class Micron(_Unit):
    """Micron waveunit."""
    name = 'um'

    @staticmethod
    def to(waveunit):
        """Convert waveunit.

        ====== ============ ======================
        From   To           Conversion
        ====== ============ ======================
        ``um`` ``m``        :math:`10^{-6}` ``um``
        ``um`` ``um``       :math:`1`
        ``um`` ``nm``       :math:`10^{3}` ``um``
        ``um`` ``angstrom`` :math:`10^{4}` ``um``
        ====== ============ ======================

        Parameters
        ----------
        waveunit : str
            Wavelength units, as accepted by :func:`Unit`.

        Returns
        -------
        int
            :attr:`wave` scaling factor

        """
        if waveunit.lower() in ['m', 'meter']:
            return 1e-6
        elif waveunit.lower() in ['um', 'micron']:
            return 1
        elif waveunit.lower() in ['nm', 'nanometer']:
            return 1e3
        elif waveunit.lower() == 'angstrom':
            return 1e4
        else:
            raise ValueError('Unknown waveunit: ' + waveunit)


class Nanometer(_Unit):
    name = 'nm'

    @staticmethod
    def to(waveunit):
        if waveunit.lower() in ['m', 'meter']:
            return 1e-9
        elif waveunit.lower() in ['um', 'micron']:
            return 1e-3
        elif waveunit.lower() in ['nm', 'nanometer']:
            return 1
        elif waveunit.lower() == 'angstrom':
            return 1e1
        else:
            raise ValueError('Unknown waveunit: ' + waveunit)


class Angstrom(_Unit):
    name = 'angstrom'

    @staticmethod
    def to(waveunit):
        if waveunit.lower() in ['m', 'meter']:
            return 1e-10
        elif waveunit.lower() in ['um', 'micron']:
            return 1e-4
        elif waveunit.lower() in ['nm', 'nanometer']:
            return 1e-1
        elif waveunit.lower() == 'angstrom':
            return 1
        else:
            raise ValueError('Unknown waveunit: ' + waveunit)


class Photlam(_Unit):
    name = 'photlam'
    longname = 'photons s^-1 m^-2'

    @staticmethod
    def to(flux, fluxunit, wave):
        if fluxunit.lower() == 'wlam':
            return flux * (H*C)/wave
        elif fluxunit.lower() == 'flam':
            return flux * (H*C)/wave * 1e7 * 1e-4
        elif fluxunit.lower() == 'photlam':
            return flux
        else:
            raise ValueError('Unknown fluxunit: ' + fluxunit)


class Flam(_Unit):
    name = 'flam'
    longname = 'erg s^-1 cm^-2'

    @staticmethod
    def to(flux, fluxunit, wave):
        if fluxunit.lower() == 'photlam':
            return flux * wave/(H * C) * 1e-7 * 1e4
        elif fluxunit.lower() == 'wlam':
            return flux * 1e-7 * 1e4
        elif fluxunit.lower() == 'flam':
            return flux
        else:
            raise ValueError('Unknown fluxunit: ' + fluxunit)


class Wlam(_Unit):
    name = 'wlam'
    longname = 'W m^-2'

    @staticmethod
    def to(flux, fluxunit, wave):
        if fluxunit.lower() == 'photlam':
            return flux * wave/(H*C)
        elif fluxunit.lower() == 'flam':
            return flux * 1e7 * 1e-4
        elif fluxunit.lower() == 'wlam':
            return flux
        else:
            raise ValueError('Unknown fluxunit: ' + fluxunit)



import numpy as np

__all__ = ['normalize_power', 'iterable_amplitude', 'iterable_mask', 'iterable_phase',
           'iterable_segmask']


def normalize_power(array, power=1):
    r"""Normalizie the power in an array.

    The total power in an array is given by

    .. math::

        P = \sum{\left|\mbox{array}\right|^2}

    A normalization coefficient is computed as

    .. math::

        c = \sqrt{\frac{p}{\sum{\left|\mbox{array}\right|^2}}}

    The array returned by a will be scaled by the normalization coefficient so
    that its power is equal to :math:`p`.

    Parameters
    ----------
    array : array_like
        Array to be normalized

    power : float, optional
        Desired power in normalized array. Default is 1.

    Returns
    -------
    array : ndarray
        Normalized array

    """
    array = np.asarray(array)
    return array * np.sqrt(power/np.sum(np.abs(array)**2))


def iterable_amplitude(iterable):
    """Construct a common amplitude array from an iterable of planes

    Parameters
    ----------
    iterable : list_like
        List of planes

    Returns
    -------
    amplitude : ndarray

    """
    amplitude = np.array(1)
    for plane in iterable:
        amplitude = amplitude * plane.amplitude
    return amplitude


def iterable_phase(iterable):
    """Construct a common phase array from an iterable of planes

    Parameters
    ----------
    iterable: list_like
        List of planes

    Returns
    -------
    phase: ndarray

    """

    phase = np.array(0)
    for plane in iterable:
        phase = phase + plane.phase
    return phase


def iterable_mask(iterable):
    """Construct a common mask array from an iterable of planes

    Parameters
    ----------
    iterable: list_like
        List of planes

    Returns
    -------
    mask: ndarray

    """
    mask = np.array(1)
    for plane in iterable:
        mask = mask * plane.mask
    return mask


def iterable_segmask(iterable):
    """Construct a common segmask array from an iterable of planes

    Parameters
    ----------
    iterable: list_like
        List of planes

    Returns
    -------
    segmask: ndarray

    """
    if all(plane.segmask is None for plane in iterable):
        return None
    else:
        segmask = np.array(1)
        for plane in iterable:
            if plane.segmask is not None:
                segmask = segmask * plane.segmask
        return segmask



import numpy as np

__all__ = ['pixel', 'jitter', 'smear']


def pixel(img, oversample=1):
    """Apply the aperture effects of a square pixel on a discretely sampled
    image.

    Parameters
    ----------
    img : array_like
        Input image

    oversample : int, optional
        Oversampling factor of img. Default is 1.

    Returns
    -------
    out : ndarray
        Image with pixel sampling effects applied.

    Example
    -------
    Apply pixel MTF to a 3x oversampled PSF:

    .. code:: pycon

        >>> import lentil
        >>> import matplotlib.pyplot as plt
        >>> psf = ...  # PSF calculation details omitted
        >>> psf_mtf = lentil.convolvable.pixel(psf, oversample=3)
        >>> psf_detector = lentil.util.rescale(psf_mtf, 1/3)

    Note that both the pixel MTF and detector resampling operations preserve
    radiometry:

    .. code:: pycon

        >>> print(np.sum(psf), np.sum(psf_mtf), np.sum(psf_detector))
        50398.80556524441 50398.80556524441 50398.80556524441

    See Also
    --------
    :func:`lentil.detector.pixelate`

    References
    ----------
    [1] https://en.wikipedia.org/wiki/Convolution_theorem

    """

    img = np.asarray(img)
    x = np.fft.fftfreq(img.shape[1])
    y = np.fft.fftfreq(img.shape[0])

    mtf_x = np.sinc(x*oversample)
    mtf_y = np.sinc(y*oversample)
    kernel = np.dot(mtf_x[:, np.newaxis], mtf_y[np.newaxis, :])

    return np.abs(np.fft.ifft2(np.fft.fft2(img)*kernel))


def jitter(img, scale, pixelscale=1, oversample=1):
    """Apply image jitter via convolution.

    Parameters
    ----------
    img : array_like
        Input image

    scale : float
        1-sigma jitter motion

        If ``pixelscale = 1``, ``scale`` should give jitter in terms of
        fractional pixels. If ``pixelscale`` is the physical dimension of
        one pixel, ``scale`` should give jitter in terms of the same units.

    pixelscale : float, optional
        Pixel size. If ``pixelscale = 1`` (default), the ``scale`` parameter
        provided when Jitter is called should give jitter in terms of fractional
        pixels. If ``pixelscale`` is the physical dimension of one pixel, ``the
        ``scale`` parameter provided when Jitter is called should give jitter in
        terms of the same units.

    oversample : int, optional
        Oversampling factor of img. Default is 1.

    Returns
    -------
    out : ndarray
        Image with jitter applied.

    Examples
    --------
    Apply 2 pixels jitter to a natively-sampled PSF:

    .. code:: pycon

        >>> import lentil
        >>> import matplotlib.pyplot as plt
        >>> psf = ...  # PSF calculation details omitted
        >>> psf_jitter = lentil.convolvable.jitter(psf, scale=2)
        >>> plt.subplot(121), plt.imshow(psf)
        >>> plt.subplot(122), plt.imshow(psf_jitter)

    .. image:: /_static/img/api/convolvable/jitter_px.png
        :width: 500px

    Apply 20 um jitter to a 3x oversampled PSF. Note that because we are
    specifying jitter in terms of linear distance on the focal plane, we must
    provide the detector pixelscale when creating the Jitter object. We also
    provide the oversampling factor, ensuring the convolution kernel is
    appropriately sized:

    .. code:: pycon

        >>> import lentil
        >>> import matplotlib.pyplot as plt
        >>> psf = ...  # PSF calculation details omitted
        >>> psf_jitter = lentil.convolvable.jitter(psf,
        ...                                        scale=20e-6,
        ...                                        pixelscale=5e-6,
        ...                                        oversample=3)
        >>> plt.subplot(121), plt.imshow(psf)
        >>> plt.subplot(122), plt.imshow(psf_jitter)

    .. image:: /_static/img/api/convolvable/jitter_m.png
        :width: 500px

    References
    ----------
    [1] https://en.wikipedia.org/wiki/Convolution_theorem

    """
    img = np.asarray(img)
    x = np.fft.fftfreq(img.shape[1])
    y = np.fft.fftfreq(img.shape[0])
    xx, yy = np.meshgrid(x, y)
    rho = np.sqrt(xx ** 2 + yy ** 2)
    kernel = np.exp(-2 * (np.pi * (scale / pixelscale) * oversample * rho) ** 2)

    out = np.abs(np.fft.ifft2(np.fft.fft2(img)*kernel))
    return out * np.sum(img) / np.sum(out)  # rescale to preserve input weight


def smear(img, distance, angle=None, pixelscale=1, oversample=1):
    """Apply image smear via convolution.

    Parameters
    ----------
    img : array_like
        Input image

    distance : float
        Linear smear distance

        If ``pixelscale = 1``, ``distance`` should give smear in terms of
        fractional pixels. If ``pixelscale`` is the physical dimension of
        one pixel, ``distance`` should give smear in terms of the same
        units.

    angle : float, optional
        Smear direction in degrees measured clockwise from the x-axis. If
        ``None`` (default), a new direction is randomly chosen every time
        :class:`Smear` is called.

    pixelscale : float, optional
        Pixel size. If ``pixelscale = 1`` (default), the ``scale`` parameter
        provided when Smear is called should give jitter in terms of fractional
        pixels. If ``pixelscale`` is the physical dimension of one pixel, ``the
        ``scale`` parameter provided when Smear is called should give jitter in
        terms of the same units.

    oversample : int, optional
        Oversampling factor of img. Default is 1.

    Returns
    -------
    out : ndarray
        Image with smear applied.

    Examples
    --------
    Apply 100 um smear at 60 degrees to a natively-sampled PSF. Note that
    because we are specifying smear in terms of linear distance on the focal
    plane, we must provide the detector pixelscale when creating the Smear
    object:

    .. code:: pycon

        >>> import lentil
        >>> import matplotlib.pyplot as plt
        >>> psf = ...  # PSF calculation details omitted
        >>> psf_smear = lentil.convolvable.smear(psf,
        ...                                      distance=100e-6,
        ...                                      angle=30,
        ...                                      pixelscale=5e-6)
        >>> plt.subplot(121), plt.imshow(psf)
        >>> plt.subplot(122), plt.imshow(psf_smear)

    .. image:: /_static/img/api/convolvable/smear_m.png
        :width: 500px

    Apply 10 pixels smear at a random angle to a 3x oversampled PSF:

    .. code:: pycon

        >>> import lentil
        >>> import matplotlib.pyplot as plt
        >>> psf = ...  # PSF calculation details omitted
        >>> psf_smear = lentil.convolvable.smear(psf, distance=10,
        ...                                      oversample=3)
        >>> plt.subplot(121), plt.imshow(psf)
        >>> plt.subplot(122), plt.imshow(psf_smear)

    .. image:: /_static/img/api/convolvable/smear_px.png
        :width: 500px

    References
    ----------
    [1] https://en.wikipedia.org/wiki/Convolution_theorem

    """
    img = np.asarray(img)
    x = np.fft.fftfreq(img.shape[1])
    y = np.fft.fftfreq(img.shape[0])

    xx, yy = np.meshgrid(x, y)

    if angle is None:
        angle = np.random.uniform(0, 2 * np.pi)
    else:
        angle = np.radians(angle)

    yy_rot = -np.sin(angle) * yy + np.cos(angle) * xx

    kernel = np.sinc(yy_rot * (distance / pixelscale) * oversample)

    out = np.abs(np.fft.ifft2(np.fft.fft2(img)*kernel))
    return out * np.sum(img) / np.sum(out)  # rescale to preserve input weight




import functools
import numpy as np

from lentil import util

__all__ = ['dft2', 'idft2', 'czt2', 'iczt2']


FFTW_WISDOM_FILENAME = '.lentil_fftw_wisdom'
_fftw_wisdom_file = lambda: os.path.join(os.path.expanduser('~'), FFTW_WISDOM_FILENAME)

def _fftw_save_wisdom():
    import pyfftw, pickle, os
    pickle.dump(pyfftw.export_wisdom(), open(_fftw_wisdom_file(), 'wb'))

try:
    import pyfftw, atexit, pickle, os
    pyfftw.interfaces.cache.enable()
    pyfftw.interfaces.cache.set_keepalive_time(5000)
    if os.path.isfile(_fftw_wisdom_file()):
        pyfftw.import_wisdom(pickle.load(open(_fftw_wisdom_file(), 'rb')))
    atexit.register(_fftw_save_wisdom)
except ImportError:
    pyfftw = None


def dft2(f, alpha, npix=None, shift=(0, 0), offset=(0, 0), unitary=True, out=None):
    """Compute the 2-dimensional discrete Fourier Transform.

    This function allows independent control over input shape, output shape,
    and output sampling by implementing the matrix triple product algorithm
    described in [1].

    Parameters
    ----------
    f : array_like
        2D array to Fourier Transform

    alpha : float or array_like
        Output plane sampling interval (frequency). If :attr:`alpha` is an
        array, ``alpha[1]`` represents row-wise sampling and ``alpha[2]``
        represents column-wise sampling. If :attr:`alpha` is a scalar,
        ``alpha[1] = alpha[2] = alpha`` gives uniform sampling across the rows
        and columns of the output plane.

    npix : int or array_like, optional
        Size of the output array :attr:`F`. If :attr:`npix` is an array,
        ``F.shape = (npix[1], npix[2])``. If :attr:`npi`` is a scalar,
        ``F.shape = (npix, npix)``. Default is ``f.shape``.

    shift : array_like, optional
        Number of pixels in (x,y) to shift the DC pixel in the output plane
        with the origin centrally located in the plane. Default is ``(0,0)``.

    offset : array_like, optional
        Number of pixels in (x,y) that the input plane is shifted relative to
        the origin. Default is ``(0,0)``.

    unitary : bool, optional
        Normalization flag. If ``True``, a normalization is performed on the
        output such that the DFT operation is unitary and energy is conserved
        through the Fourier transform operation (Parseval's theorem). In this
        way, the energy in in a limited-area DFT is a fraction of the total
        energy corresponding to the limited area. Default is ``True``.

    out : ndarray or None
        A location into which the result is stored. If provided, it must have
        shape = npix and dtype = np.complex. If not provided or None, a
        freshly-allocated array is returned.

    Returns
    -------
    F : complex ndarray

    Notes
    -----
    * Setting ``alpha = 1/f.shape`` and ``npix = f.shape`` is equivalent to

      ::

          F = np.fft.ifftshift(np.fft.fft2(np.fft.fftshift(f)))

    * ``dft2()`` is designed to place the DC pixel in the same location as a
      well formed call to any standard FFT for both even and odd sized input
      arrays. The DC pixel is located at ``np.floor(npix/2) + 1``, which is
      consistent with calls to Numpy's FFT method where the input and output
      are correctly shifted:
      ``np.fft.ifftshift(np.fft.fft2(np.fft.fftshift(f)))``.

    * If the y-axis shift behavior is not what you are expecting, you most
      likely have your plotting axes flipped (matplotlib's default behavior is
      to place [0,0] in the upper left corner of the axes). This may be resolved
      by either flipping the sign of the y component of ``shift`` or by passing
      ``origin = 'lower'`` to ``imshow()``.

    References
    ----------
    [1] Soummer, et. al. Fast computation of Lyot-style coronagraph propagation (2007)

    """

    ax, ay = _sanitize_alpha(alpha)

    f = np.asarray(f)
    m, n = f.shape

    if npix is None:
        npix = [m, n]
    M, N = _sanitize_npix(npix)

    if offset is None:
        offset = (0, 0)

    sx, sy = _sanitize_shift(shift)
    ox, oy = _sanitize_npix(offset)

    if out is not None:
        if not np.can_cast(complex, out.dtype):
            raise TypeError(f"Cannot cast complex output to dtype('{out.dtype}')")

    E1, E2 = _dft2_matrices(m, n, M, N, ax, ay, sx, sy, ox, oy)
    F = np.dot(E1.dot(f), E2, out=out)

    # now calculate the answer, without reallocating memory
    if unitary:
        np.multiply(F, np.sqrt(np.abs(ax * ay)), out=F)

    return F


@functools.lru_cache(maxsize=32)
def _dft2_matrices(m, n, M, N, ax, ay, sx, sy, ox, oy):
    X, Y, U, V = _dft2_coords(m, n, M, N)
    E1 = util.expc(-2.0 * np.pi * ay * np.outer(Y-sy+oy, V-sy)).T
    E2 = util.expc(-2.0 * np.pi * ax * np.outer(X-sx+ox, U-sx))
    return E1, E2


@functools.lru_cache(maxsize=32)
def _dft2_coords(m, n, M, N):
    # Y and X are (r,c) coordinates in the (m x n) input plane f
    # V and U are (r,c) coordinates in the (M x N) ourput plane F

    X = np.arange(n) - np.floor(n/2.0)
    Y = np.arange(m) - np.floor(m/2.0)
    U = np.arange(N) - np.floor(N/2.0)
    V = np.arange(M) - np.floor(M/2.0)

    return X, Y, U, V


def idft2(F, alpha, npix=None, shift=(0,0), unitary=True, out=None):
    """Compute the 2-dimensional inverse discrete Fourier Transform.

    This function allows independent control over input shape, output shape,
    and output sampling by implementing the matrix triple product algorithm
    described in [1].

    Parameters
    ----------
    F : array_like
        2D array to Fourier Transform

    alpha : float or array_like
        Input plane sampling interval (frequency). If :attr:`alpha` is an array,
        ``alpha[1]`` represents row-wise sampling and ``alpha[2]`` represents
        column-wise sampling. If :attr:`alpha` is a scalar,
        ``alpha[1] = alpha[2] = alpha`` represents uniform sampling across the
        rows and columns of the input plane.

    npix : int or array_like, optional
        Size of the output array :attr:`F`. If :attr:`npix` is an array,
        ``F.shape = (npix[1], npix[2])``. If :attr:`npix` is a scalar,
        ``F.shape = (npix, npix)``. Default is ``F.shape``

    shift : array_like, optional
        Number of pixels in (x,y) to shift the DC pixel in the output plane with
        the origin centrally located in the plane. Default is `[0,0]`.

    unitary : bool, optional
        Normalization flag. If ``True``, a normalization is performed on the
        output such that the DFT operation is unitary and energy is conserved
        through the Fourier transform operation (Parseval's theorem). In this
        way, the energy in in a limited-area DFT is a fraction of the total
        energy corresponding to the limited area. Default is ``True``.

    Returns
    -------
    f : complex ndarray

    Notes
    -----
    * Setting ``alpha = 1/F.shape`` and ``npix = F.shape`` is equivalent to

      ::

          F = np.fft.ifftshift(np.fft.ifft2(np.fft.fftshift(F)))

    * ``idft2()`` is designed to place the DC pixel in the same location as a
      well formed call to any standard FFT for both even and odd sized input
      arrays. The DC pixel is located at ``np.floor(npix/2) + 1``, which is
      consistent with calls to Numpy's IFFT method where the input and output
      are correctly shifted:
      ``np.fft.ifftshift(np.fft.ifft2(np.fft.fftshift(f)))``.

    * If the y-axis shift behavior is not what you are expecting, you most
      likely have your plotting axes flipped (matplotlib's default behavior is
      to place [0,0] in the upper left corner of the axes). This may be resolved
      by either flipping the sign of the y component of ``shift`` or by passing
      ``origin = 'lower'`` to ``imshow()``.

    References
    ----------
    [1] Soummer, et. al. Fast computation of Lyot-style coronagraph propagation (2007)

    [2] `Expressing the inverse DFT in terms of the DFT <https://en.wikipedia.org/wiki/Discrete_Fourier_transform#Expressing_the_inverse_DFT_in_terms_of_the_DFT>`_.

    """
    F = np.asarray(F)
    N = F.size
    # will allocate memory for F if out == None
    F = dft2(np.conj(F), alpha, npix, shift, unitary=unitary, out=out)
    np.conj(F, out=F)
    return np.divide(F, N, out=F)



def czt2(f, alpha, npix=None, shift=(0, 0), unitary=True):
    """Compute the 2-dimensional discrete Fourier Transform using the chirp
    z-transform.

    Parameters
    ----------
    f : array_like
        2D array to Fourier Transform

    alpha : float or array_like
        Output plane sampling interval (frequency). If :attr:`alpha` is an
        array, ``alpha[1]`` represents row-wise sampling and ``alpha[2]``
        represents column-wise sampling. If :attr:`alpha` is a scalar,
        ``alpha[1] = alpha[2] = alpha`` gives uniform sampling across the rows
        and columns of the output plane.

    npix : int or array_like, optional
        Size of the output array :attr:`F`. If :attr:`npix` is an array,
        ``F.shape = (npix[1], npix[2])``. If :attr:`npi`` is a scalar,
        ``F.shape = (npix, npix)``. Default is ``f.shape``.

    shift : array_like, optional
        Number of pixels in (x,y) to shift the DC pixel in the output plane
        with the origin centrally located in the plane. Default is ``[0,0]``.

    unitary : bool, optional
        Normalization flag. If ``True``, a normalization is performed on the
        output such that the DFT operation is unitary and energy is conserved
        through the Fourier transform operation (Parseval's theorem). In this
        way, the energy in in a limited-area DFT is a fraction of the total
        energy corresponding to the limited area. Default is ``True``.

    Returns
    -------
    F : complex ndarray

    Notes
    -----
    * If ``pyfftw`` is available it is used to increase FFT performance

    * Setting ``alpha = 1/f.shape`` and ``npix = f.shape`` is equivalent to

      ::

          F = np.fft.ifftshift(np.fft.fft2(np.fft.fftshift(f)))

    * ``czt2()`` is designed to place the DC pixel in the same location as a
      well formed call to any standard FFT for both even and odd sized input
      arrays. The DC pixel is located at ``np.floor(npix/2) + 1``, which is
      consistent with calls to Numpy's FFT method where the input and output
      are correctly shifted:
      ``np.fft.ifftshift(np.fft.fft2(np.fft.fftshift(f)))``.

    * If the y-axis shift behavior is not what you are expecting, you most
      likely have your plotting axes flipped (matplotlib's default behavior is
      to place [0,0] in the upper left corner of the axes). This may be resolved
      by either flipping the sign of the y component of ``shift`` or by passing
      ``origin = 'lower'`` to ``imshow()``.

    References
    ----------
    [1] Rabiner, et. al. The Chirp z-Transform Algorithm. Iransactions on Audio and Electroacoustics. (1969)

    """
    ax, ay = _sanitize_alpha(alpha)

    f = np.asarray(f)
    m, n = f.shape

    if npix is None:
        npix = [m, n]
    M, N = _sanitize_npix(npix)

    sx, sy = _sanitize_shift(shift)

    L = _czt2_optimal_L((m,n), (M,N))

    z_row, beta_row, gamma_row = _czt_coeffs(m, M, L[0], sy, ay, unitary)
    z_col, beta_col, gamma_col = _czt_coeffs(n, N, L[1], sx, ax, unitary)

    f_phased = np.multiply(f, beta_col)
    f_phased = np.multiply(f_phased, beta_row[:, np.newaxis], out=f_phased)

    if pyfftw:
        H = pyfftw.interfaces._utils._Xfftn(f_phased, L, (-2,-1), False, 'FFTW_MEASURE',
                                           48, True, True, 'fft2', normalise_idft=False,
                                           ortho=False)
        H = np.multiply(H, z_col, out=H)
        H = np.multiply(z_row[:, np.newaxis], H, out=H)
        X = pyfftw.interfaces._utils._Xfftn(H, L, (-2,-1), False, 'FFTW_MEASURE', 48,
                                           True, True, 'ifft2', normalise_idft=False,
                                           ortho=False)
    else:
        H = np.fft.fft2(f_phased, L)
        H = np.multiply(H, z_col, out=H)
        H = np.multiply(z_row[:, np.newaxis], H, out=H)
        X = np.fft.ifft2(H)

    X = np.multiply(X[0:M, 0:N], gamma_col)
    X = np.multiply(gamma_row[:, np.newaxis], X, out=X)
    return X


def iczt2(F, alpha, npix=None, shift=(0, 0), unitary=True):
    """Compute the 2-dimensional inverse discrete Fourier Transform using the
    chirp z-transform.

    Parameters
    ----------
    f : array_like
        2D array to Fourier Transform

    alpha : float or array_like
        Output plane sampling interval (frequency). If :attr:`alpha` is an
        array, ``alpha[1]`` represents row-wise sampling and ``alpha[2]``
        represents column-wise sampling. If :attr:`alpha` is a scalar,
        ``alpha[1] = alpha[2] = alpha`` gives uniform sampling across the rows
        and columns of the output plane.

    npix : int or array_like, optional
        Size of the output array :attr:`F`. If :attr:`npix` is an array,
        ``F.shape = (npix[1], npix[2])``. If :attr:`npi`` is a scalar,
        ``F.shape = (npix, npix)``. Default is ``f.shape``.

    shift : array_like, optional
        Number of pixels in (x,y) to shift the DC pixel in the output plane
        with the origin centrally located in the plane. Default is ``[0,0]``.

    unitary : bool, optional
        Normalization flag. If ``True``, a normalization is performed on the
        output such that the DFT operation is unitary and energy is conserved
        through the Fourier transform operation (Parseval's theorem). In this
        way, the energy in in a limited-area DFT is a fraction of the total
        energy corresponding to the limited area. Default is ``True``.

    Returns
    -------
    F : complex ndarray

    Notes
    -----
    * If ``pyfftw`` is available it is used to increase FFT performance

    * Setting ``alpha = 1/F.shape`` and ``npix = F.shape`` is equivalent to

      ::

          f = np.fft.ifftshift(np.fft.ifft2(np.fft.fftshift(F)))

    * ``czt2()`` is designed to place the DC pixel in the same location as a
      well formed call to any standard FFT for both even and odd sized input
      arrays. The DC pixel is located at ``np.floor(npix/2) + 1``, which is
      consistent with calls to Numpy's FFT method where the input and output
      are correctly shifted:
      ``np.fft.ifftshift(np.fft.fft2(np.fft.fftshift(f)))``.

    * If the y-axis shift behavior is not what you are expecting, you most
      likely have your plotting axes flipped (matplotlib's default behavior is
      to place [0,0] in the upper left corner of the axes). This may be resolved
      by either flipping the sign of the y component of ``shift`` or by passing
      ``origin = 'lower'`` to ``imshow()``.

    References
    ----------
    [1] Rabiner, et. al. The Chirp z-Transform Algorithm. Iransactions on Audio and Electroacoustics. (1969)

    [2] `Expressing the inverse DFT in terms of the DFT <https://en.wikipedia.org/wiki/Discrete_Fourier_transform#Expressing_the_inverse_DFT_in_terms_of_the_DFT>`_.

    """
    F = np.asarray(F)
    N = F.size
    # will allocate memory for F if out == None
    F = czt2(np.conj(F), alpha, npix, shift, unitary=unitary, out=out)
    np.conj(F, out = F)
    return np.divide(F, F.size,out=F)


@functools.lru_cache(maxsize=32, typed=True)
def _czt_coeffs(N, M, L, shift, alpha, unitary):

    dN = N // 2
    dM = M // 2 + shift

    m_hat = np.linspace(start=-dM, stop=M-dM-1, num=M)

    beta = _czt_beta(N, dN, L, alpha, unitary)

    gamma = util.expc(-np.pi * np.square(m_hat) * alpha)


    z_hat_i = np.zeros(L)
    start = dN - dM

    p = np.linspace(start=start, stop=start+M-1, num=M)
    z_hat_i[0:M] = np.pi * np.square(p)

    p = np.linspace(start=start-N+1, stop=start-1, num=N-1)
    z_hat_i[L-N+1:L] = np.pi * np.square(p)

    z_hat = util.expc(z_hat_i * alpha)
    z_hat[M:L-N+1] = 0
    z = np.fft.fft(z_hat)

    return z, beta, gamma

@functools.lru_cache(maxsize=32, typed=True)
def _czt_beta(N, dN, L, alpha, unitary):
    n_hat = np.linspace(start=-dN, stop=N-dN-1, num=N)

    if unitary:
        scale_factor = np.sqrt(alpha)
    else:
        scale_factor = 1

    if pyfftw:
        scale_factor /= L

    return scale_factor * util.expc(-np.pi * np.square(n_hat) * alpha)

@functools.lru_cache(maxsize=32, typed=True)
def _czt2_optimal_L(N, M):
    divisors = set([2, 3, 5, 7]) # this is what FFTW is fast at
    L = []
    for k in range(2):
        min_L = N[k] + M[k] - 1

        for candidate_L in range(min_L, 2 * min_L + 1):
            current_L = candidate_L
            rem = 0
            while rem == 0:
                for div in divisors:
                    rem = current_L % div
                    if not rem: # div evenly divides current_L
                        current_L = current_L//div
                        if current_L in divisors:
                            L.append(int(candidate_L))
                            rem = 1 # exit the while loop
                        break
            if len(L) == k + 1:
                break

    return tuple(L)


def _sanitize_alpha(x):
    """Return consistent representation of alpha as ax, ay"""
    x = np.asarray(x)
    if x.size == 1:
        ay, ax = float(x), float(x)
    else:
        ay, ax = float(x[0]), float(x[1])
    return ax, ay


def _sanitize_npix(x):
    """Return consistent representation of npix as M, N"""
    x = np.asarray(x)
    if x.size == 1:
        M, N = int(x), int(x)
    else:
        M, N = int(x[0]), int(x[1])
    return M, N


def _sanitize_shift(x):
    """Return consistent representation of shift as sx, sy"""
    if isinstance(x, np.ndarray):
        sx, sy = float(x[0]), float(x[1])
    else:
        sx, sy = x
    return sx, sy



from lentil import convolvable
from lentil import detector
from lentil import modeltools
from lentil.plane import *
from lentil.prop import propagate
from lentil import radiometry
from lentil import util
from lentil import wfe
from lentil import zernike

__version__ = '0.5.0b4'



from copy import deepcopy

import numpy as np
from scipy import ndimage
import scipy.integrate
import scipy.optimize

from lentil import util

__all__ = ['Plane', 'Pupil', 'Image', 'Detector', 'DispersiveShift', 'Grism',
           'LensletArray', 'Tilt', 'Rotate', 'Flip']


class Plane:
    """Base class for representing a finite geometric plane.

    Parameters
    ----------
    pixelscale : float or (2,) array_like
        Physical sampling (in meters) of each pixel in the plane. If :attr:`pixelscale`
        is a scalar, uniform sampling in x and y is assumed.

    amplitude : array_like, optional
        Electric field amplitude transmission. Amplitude should be normalized
        with :func:`~lentil.util.normalize_power` if conservation of power
        through a diffraction propagation is required. If not specified, a
        default amplitude is created which has no effect on wavefront
        propagation.

    phase : array_like, optional
        Phase change caused by plane. If not specified, a default phase is
        created which has no effect on wavefront propagation.

    mask : array_like, optional
        Binary mask. If not specified, a mask is created from the amplitude.

    segmask : (nseg, ...) {array_like, None} optional
        Binary segment mask. Default is None.

    Attributes
    ----------
    tilt : list

    cache : dict

    """

    def __init__(self, pixelscale=None, amplitude=1, phase=0, mask=None, segmask=None):
        # We directly set the local attributes here in case a subclass has redefined
        # the property (which could cause an weird behavior and will throw an
        # AttributeError if the subclass hasn't defined an accompanying getter
        self.pixelscale = pixelscale
        self._amplitude = np.asarray(amplitude) if amplitude is not None else None
        self._phase = np.asarray(phase) if phase is not None else None
        self._mask = np.asarray(mask) if mask is not None else None
        self._segmask = np.asarray(segmask) if segmask is not None else None

        self.cache = {}
        self.tilt = []

    def __init_subclass__(cls):
        # we have to define default values to avoid AttributeErrors in case
        # subclasses don't call super().__init__ and don't explicitly define
        # an attribute
        cls._pixelscale = None
        cls._amplitude = np.array(1)
        cls._phase = np.array(0)
        cls._mask = None
        cls._segmask = None

        cls.cache = {}
        cls.tilt = []

    def __repr__(self):
        return f'{self.__class__.__name__}()'

    @property
    def pixelscale(self):
        """Physical sampling (in meters) of each pixel in the plane.

        Returns
        -------
        pixelscale : ndarray
            (dx,dy) sampling of the Plane
        """
        return self._pixelscale

    @pixelscale.setter
    def pixelscale(self, value):
        if value is not None:
            value = np.asarray(value)
            if value.shape == ():
                value = np.append(value, value)
        self._pixelscale = value

    @property
    def amplitude(self):
        """Electric field amplitude transmission.

        Returns
        -------
        amplitude : ndarray

        Note
        ----
        If a cached value exists it is returned first.

        """
        return self._amplitude

    @amplitude.setter
    def amplitude(self, value):
        if value is not None:
            self._amplitude = np.asarray(value)
        else:
            self._amplitude = None

    @property
    def phase(self):
        """Electric field phase shift.

        Returns
        -------
        phase : ndarray

        Note
        ----
        If a cached value exists it is returned first.

        """
        return self._phase

    @phase.setter
    def phase(self, value):
        if value is not None:
            self._phase = np.asarray(value)
        else:
            self._phase = None

    @property
    def mask(self):
        """Binary mask

        Returns
        -------
        mask : ndarray

        """
        if self._mask is not None:
            return self._mask
        else:
            if self.amplitude is not None:
                mask = np.copy(self.amplitude)
                mask[mask != 0] = 1
                return mask
            else:
                return None

    @mask.setter
    def mask(self, value):
        if value is not None:
            self._mask = np.asarray(value)
        else:
            self._mask = None

    @property
    def segmask(self):
        """Binary segment mask

        Returns
        -------
        segmask : (nseg,...) ndarray or None

        """
        return self._segmask

    @segmask.setter
    def segmask(self, value):
        if value is None:
            segmask = None
        else:
            # Note this operation implicitly converts a list or tuple of arrays
            # to an appropriately stacked ndarray
            segmask = np.asarray(value)
            if segmask.ndim == 2:
                segmask = segmask[np.newaxis, ...]
            elif segmask.ndim > 3:
                raise ValueError(f"don't know what to do with {segmask.ndim}th dimension")
        self._segmask = segmask

    @property
    def nseg(self):
        """Number of segments represented in the plane."""
        if self.segmask is None:
            return 0
        else:
            return self.segmask.shape[0]

    @property
    def shape(self):
        """Plane dimensions computed from :attr:`mask` or :attr:`segmask`. Returns None
        if :attr:`mask` and :attr:`segmask` are None.
        """
        if self.mask is None and self.segmask is None:
            return None
        else:
            if self.mask is not None:
                return self.mask.shape
            else:
                return self.segmask.shape[1], self.segmask.shape[2]

    @staticmethod
    def rescale_amplitude(amplitude, scale):
        """Rescale plane amplitude

        This method uses 3rd order interpolation and preserves the power of the amplitude
        array. If custom interpolation is required, this method should be redefined.

        Parameters
        ----------
        amplitude : array_like

        scale : float
            Scaling factor. Scale factors less than 1 will shrink the amplitude. Scale
            factors greater than 1 will grow the amplitude.

        Returns
        -------
        amplitude : ndarray
            Rescaled amplitude

        """
        if scale == 1:
            return amplitude
        else:
            return util.rescale(amplitude, scale=scale, shape=None, mask=None,
                                order=3, mode='nearest', unitary=False)/scale**2

    @staticmethod
    def rescale_phase(phase, scale):
        """Rescale plane phase

        This method uses 3rd order interpolation. If custom interpolation is required,
        this method should be redefined.

        Parameters
        ----------
        scale : float
            Scaling factor. Scale factors less than 1 will shrink the amplitude. Scale
            factors greater than 1 will grow the amplitude.

        Returns
        -------
        phase : ndarray
            Rescaled phase

        """
        if scale == 1:
            return phase
        else:
            return util.rescale(phase, scale=scale, shape=None, mask=None,
                                order=3, mode='nearest', unitary=False)

    @staticmethod
    def rescale_mask(mask, scale):
        """Rescale plane mask

        This method uses linear interpolation. If custom interpolation is required,
        this method should be redefined.

        Parameters
        ----------
        scale : float
            Scaling factor. Scale factors less than 1 will shrink the amplitude. Scale
            factors greater than 1 will grow the amplitude.

        Returns
        -------
        mask : ndarray
            Rescaled mask

        """
        mask = util.rescale(mask, scale=scale, shape=None, mask=None, order=0,
                            mode='constant', unitary=False)
        #mask[mask < np.finfo(mask.dtype).eps] = 0
        mask[np.nonzero(mask)] = 1
        return mask.astype(np.int)

    @staticmethod
    def rescale_segmask(segmask, scale):
        """Rescale plane segmask

        This method uses linear interpolation. If custom interpolation is required,
        this method should be redefined.

        Parameters
        ----------
        scale : float
            Scaling factor. Scale factors less than 1 will shrink the amplitude. Scale
            factors greater than 1 will grow the amplitude.

        Returns
        -------
        segmask : ndarray
            Rescaled segmask

        """
        out = []
        for seg in segmask:
            mask = util.rescale(seg, scale=scale, shape=None, mask=None,
                                order=1, mode='nearest', unitary=False)
            mask[mask != 0] = 1
            out.append(mask)
        return np.asarray(out)

    @property
    def ptt_vector(self):
        """2D vector representing piston and tilt in x and y. Planes with no mask or
        segmask have ptt_vector = None.

        Returns
        -------
        ptt_vector : ndarray or None

        """

        # if there's no mask, we just set ptt_vector to None and move on
        if (self.shape == () or self.shape is None) and self.segmask is None:
            ptt_vector = None
        else:
            # compute unmasked piston, tip, tilt vector
            x, y = util.mesh(self.shape)
            unmasked_ptt_vector = np.einsum('ij,i->ij', [np.ones(x.size), x.ravel(), y.ravel()],
                                            [1, self.pixelscale[0], self.pixelscale[1]])

            if self.segmask is None:
                ptt_vector = np.einsum('ij,j->ij', unmasked_ptt_vector, self.mask.ravel())
            else:
                # prepare empty ptt_vector
                ptt_vector = np.empty((self.nseg * 3, self.mask.size))

                # loop over the segments and fill in the masked ptt_vectors
                for seg in np.arange(self.nseg):
                    ptt_vector[3 * seg:3 * seg + 3] = unmasked_ptt_vector * self.segmask[seg].ravel()

        return ptt_vector

    def fit_tilt(self, phase=None, ptt_vector=None):
        """Fit and remove tilt from a phase via least squares.

        Parameters
        ----------
        phase : array_like or None
            Phase to fit. If None (default), the Plane's :attr:`~lentil.Plane.phase`
            attribute it used.

        ptt_vector : array_like or None
            Piston, tip, tilt basis used for fitting. If None (default), the Plane's
            :attr:`~lentil.Plane.ptt_vector` attribute it used.

        Returns
        -------
        phase_no_tilt : ndarray
            ``phase`` with tilt fit and removed. If ``len(ptt_vector)//3 > 1``, the
            returned phase array will contain one entry for each triad in ``ptt_vector``.

        tilt : list
            List of :class:`~lentil.Tilt` objects representing the fit and removed tilt.

        Note
        ----
        This method is called during :func:`~lentil.propagate` when ``tilt='angle'``. If
        custom tilt-removal behavior is required, this method should be modified in a
        subclass.

        If tilt-removal should `always` be skipped, this method should be modified in a
        subclass to: ``return self.phase, []``

        """
        if phase is None:
            phase = self.phase

        if ptt_vector is None:
            ptt_vector = self.ptt_vector

        # There are a couple of cases where we don't have enough information to remove the
        # tilt, so we just return the phase as is and None tilt
        if ptt_vector is None or phase.size == 1:
            return phase, []

        nseg = len(ptt_vector)//3

        if nseg == 1:
            t = np.linalg.lstsq(ptt_vector.T, phase.ravel(), rcond=None)[0]
            phase_tilt = np.einsum('ij,i->j', ptt_vector[1:3], t[1:3])
            phase_no_tilt = phase - phase_tilt.reshape(phase.shape)

            # 01da13b transitioned from specifying tilt in terms of image plane
            # coordinates to about the Tilt plane axes. We can transform from
            # notional image plane to Tilt plane with x_tilt = -y_img, y_tilt = x_img
            tilt = [Tilt(x=-t[2], y=t[1])]
            return phase_no_tilt, tilt
        else:
            t = np.empty((nseg, 3))
            phase_no_tilt = np.empty((nseg, phase.shape[0], phase.shape[1]))

            # iterate over the segments and compute the tilt term
            for seg in np.arange(nseg):
                t[seg] = np.linalg.lstsq(ptt_vector[3 * seg:3 * seg + 3].T, phase.ravel(),
                                         rcond=None)[0]
                seg_tilt = np.einsum('ij,i->j', ptt_vector[3 * seg + 1:3 * seg + 3], t[seg, 1:3])
                phase_no_tilt[seg] = phase - seg_tilt.reshape(phase.shape)

            # 01da13b transitioned from specifying tilt in terms of image plane
            # coordinates to about the Tilt plane axes. We can transform from
            # notional image plane to Tilt plane with x_tilt = -y_img, y_tilt = x_img
            tilt = [Tilt(x=-t[seg, 2], y=t[seg, 1]) for seg in range(nseg)]

            return phase_no_tilt, tilt

    def slice(self, mask=None):
        """Compute slices defined by the data extent in ``mask``.

        Parameters
        ----------
        mask : array_like or None, optional
            Mask to use for identifying data extents. If None, the Plane's
            :attr:`segmask` is used if available, otherwise the Plane's
            :attr:`mask` is used. If both :attr:`segmask` and :attr:`mask` are
            None, ``Ellipsis`` is returned (slice returns all data).

        Returns
        -------
        slices : list
            List of slices corresponding to the data extent defined by ``mask``.

        See also
        --------
        * :func:`~lentil.util.boundary_slice`
        * :func:`~lentil.Plane.slice_offset`

        """

        # If a mask is not provided, we should prefer segmask over mask
        if mask is None:
            if self.segmask is not None:
                mask = self.segmask
            else:
                mask = self.mask

        # self.mask and self.segmask may still return None so we
        # catch that here
        if mask.ndim < 2 or mask is None:
            # np.s_[...] = Ellipsis -> returns the whole array
            s = [np.s_[...]]
        elif mask.ndim == 2:
            s = [util.boundary_slice(mask)]
        elif mask.ndim == 3:
            s = [util.boundary_slice(segmask) for segmask in mask]
        return s

    def slice_offset(self, slices, shape=None, indexing='xy'):
        """Compute (r,c) offsets of the centers of each slice in a list of slices.

        Parameters
        ----------
        slices : list
            List of slices

        shape : (2,) array_like or None, optional
            Shape of containing array each slice is taken from. If None (default),
            ``shape = Plane.shape``.

        indexing : {'xy', 'ij'}, optional
            Offset ordering. Default is 'xy'.

        Returns
        -------
        offsets : list
           List of center offsets corresponding to the provided slices

        See also
        --------
        * :func:`~lentil.util.slice_offset`
        * :func:`~lentil.Plane.slice`

        """

        if shape is None:
            shape = self.shape

        offsets = []
        for s in slices:
            offset = util.slice_offset(slice=s, shape=shape, indexing=indexing)
            if offset:
                offsets.append(offset)
        if offsets:
            return offsets
        else:
            return [None]

    def multiply(self, wavefront):
        """Multiply with a wavefront

        Parameters
        ----------
        wavefront : :class:`~lentil.wavefront.Wavefront` object
            Wavefront to be multiplied

        tilt : {'phase', 'angle'}, optional
            * 'phase' - any tilt present in the Element phase contribution is
              included in the Element*Wavefront product. (Default)
            * 'angle' - any tilt present in the Element phase is removed before
              computing the Element*Wavefront product. The equivalent angular
              tilt is included in the Wavefront's
              :attr:`~lentil.Wavefront.tilt` attribute.

        Note
        ----
        It is possible to customize the way multiplication is performed in this
        class and any subclasses by reimplementing any of the following methods:

        * ``_multiply_phase`` is called when :attr:`tilt` is 'phase'
        * ``_multiply_angle`` is called when :attr:`tilt` is 'angle'

        In each case, the interface is defined as follows:

        ``wavefront = _multiply_method(amplitude, phase, wavefront)``

        Returns
        -------
        wavefront : :class:`~lentil.wavefront.Wavefront` object
            Updated wavefront

        """

        if wavefront.pixelscale is None:
            wavefront.pixelscale = self.pixelscale
        else:
            if self.pixelscale is not None:
                assert all(wavefront.pixelscale == self.pixelscale)

        # TODO: could move these back inside the cached attribute construct to make it
        # easier for users to redefint multiply() in subclasses - i.e. if these attributes
        # are automatically grabbed from the cache if they exist, the user won't have to
        # recreate the block of code below in a subclasses multiply() method
        amplitude = self.cache['amplitude'] if 'amplitude' in self.cache else self.amplitude
        phase = self.cache['phase'] if 'phase' in self.cache else self.phase
        tilt = self.cache['tilt'] if 'tilt' in self.cache else self.tilt
        slc = self.cache['slice'] if 'slice' in self.cache else [np.s_[...]]
        #ofst = self.cache['offset'] if 'offset' in self.cache else [None]

        data = wavefront.data  # grab a pointer to the existing wavefront.data
        wavefront.data = []

        # wavefront.offset = []
        # then below, under wavefront.data.append:
        # wavefront.offset.append(ofst[seg])
        #   OR
        # wavefront.offset.append(

        # WAIT!
        # -----------------------------
        # what if we just compute the offset on the fly here with each slice?
        #
        # no need to cache it or deal with any of this BS?!
        #
        # Still need to figure out how to deal with the fact that we may have an
        # incoming offset that may or may not be replaced by something new.
        #
        # this could be handled by a simple rule:
        #   new offset always takes prescedence over old offset UNLESS new offset is None and old offset is not none


        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # maybe enforce a simple rule
        # wavefront with offsets can only be multiplied with Planes that either introduce
        # NO offset or have the SAME offset

        offset = []

        for d in data:
            if phase.ndim == 3:

                # we should make some assertions about wavefront.offset here

                # This is the case we hit when tilt = 'angle' and nseg > 1
                for seg in np.arange(phase.shape[0]):
                    phasor = _phasor(amplitude, phase[seg], self.segmask[seg], wavefront.wavelength, slc[seg])
                    wavefront.data.append(d[slc[seg]] * phasor)

                    # we have a few rules for dealing with the slice offset here
                    ofst = util.slice_offset(slc[seg], amplitude.shape)

                    # if wavefront.offset is empty, we will just append the offset for each slice as we go (even if offset is None)
                    #if not wavefront.offset:
                    if ofst:
                        offset.append(ofst)
                    else:
                        offset.append(None)

                    # if len(wavefront.offset) == 1

                    #if len(wavefront.offset) == 1:


            else:

                # we should make some assertions about wavefront.offset here

                # tilt = 'phase' regardless of nseg or tilt = 'angle' and nseg = None
                for s in slc:
                    phasor = _phasor(amplitude, phase, self.mask, wavefront.wavelength, s)
                    wavefront.data.append(d[s] * phasor)

                    # we have a few rules for dealing with the slice offset here
                    ofst = util.slice_offset(s, amplitude.shape)

                    # if wavefront.offset is empty, we will just append the offset for each slice as we go (even if offset is None)
                    # if not wavefront.offset:
                    if ofst:
                        offset.append(ofst)
                    else:
                        offset.append(None)

        if not wavefront.offset or not all(wavefront.offset):
            # of wavefront.offset is [] or a list of [None]:
             wavefront.offset = offset

        # TODO: verify this should really be extend and not append
        wavefront.tilt.extend(tilt)

        return wavefront


def _phasor(amplitude, phase, mask, wavelength, slc=Ellipsis):
    amp = amplitude if amplitude.size == 1 else amplitude[slc]
    ph = phase if phase.size == 1 else phase[slc]
    msk = mask if mask.size == 1 else mask[slc]
    return (amp * util.expc(2*np.pi*ph/wavelength)) * msk


class Pupil(Plane):
    """Class for representing a pupil plane.

    Parameters
    ----------
    diameter : float
        Diameter in meters

    focal_length : float
        Focal length in meters

    pixelscale : float
        Physical sampling (in meters) of each pixel in the pupil

    amplitude : array_like, optional
        Electric field amplitude transmission. Amplitude should be normalized
        with :func:`~lentil.util.normalize_power` if conservation of power
        through a diffraction propagation is required. If not specified, a
        default amplitude is created which has no effect on wavefront
        propagation.

    phase : array_like, optional
        Phase change caused by plane. If not specified, a default phase is
        created which has no effect on wavefront propagation.

    mask : array_like, optional
        Binary mask. If not specified, a mask is created from the amplitude.

    segmask : (nseg, ...) {array_like, None} optional
        Binary segment mask. Default is None.

    Note
    ----
    By definition, a pupil is represented by a spherical wavefront. Any
    aberrations in the optical system appear as deviations from this perfect
    sphere. The primary use of :class:`Pupil` is to represent these aberrations

    """
    def __init__(self, diameter=None, focal_length=None, pixelscale=None, amplitude=1,
                 phase=0, mask=None, segmask=None):

        super().__init__(pixelscale=pixelscale, amplitude=amplitude, phase=phase,
                         mask=mask, segmask=segmask)

        # We directly set the local attributes here in case a subclass has redefined
        # the property (which could cause an weird behavior and will throw an
        # AttributeError if the subclass hasn't defined an accompanying getter
        self._diameter = diameter
        self._focal_length = focal_length

    def __init_subclass__(cls):
        cls._diameter = None
        cls._focal_length = None

    @property
    def diameter(self):
        """Optical system diameter in meters."""
        return self._diameter

    @diameter.setter
    def diameter(self, value):
        self._diameter = value

    @property
    def focal_length(self):
        """Optical system focal length in meters."""
        return self._focal_length

    @focal_length.setter
    def focal_length(self, value):
        self._focal_length = value

    @property
    def f_number(self):
        """F-number."""
        return self.focal_length/self.diameter

    def multiply(self, wavefront):

        wavefront = super().multiply(wavefront)

        # we inherit the plane's focal length as the wavefront's focal length
        wavefront.focal_length = self.focal_length
        wavefront.planetype = 'pupil'

        return wavefront


class Image(Plane):
    """Class for representing an image plane.

    Parameters
    ----------
    pixelscale : float, optional
        Pixel size in meters. Pixels are assumed to be square. Default is None.

    shape : {int, (2,) array_like}, optional
        Number of pixels as (rows, cols). If a single value is provided,
        :class:`Image` is assumed to be square with nrows = ncols = shape.
        Default is None.

    """

    def __init__(self, pixelscale=None, shape=None):

        super().__init__(pixelscale=pixelscale)

        # Ensure shape is a (2,) tuple or None
        if shape:
            shape = np.asarray(shape)
            if shape.size == 1:
                shape = np.append(shape, shape)
            elif shape.size > 2:
                raise IndexError('too many values')
            shape = tuple(shape)

        self._shape = shape

    @property
    def shape(self):
        """Number of pixels as (rows, cols)."""
        return self._shape

    @shape.setter
    def shape(self, value):
        self._shape = value

    def fit_tilt(self, *args, **kwargs):
        return self.phase, []

    def slice(self, *args, **kwargs):
        # np.s_[...] = Ellipsis -> returns the whole array
        return [np.s_[...]]

    def multiply(self, wavefront):
        """Multiply with a :class:`~lentil.wavefront.Wavefront`."""

        # TODO: we should construct some version of Plane.multiply() here or just call super().multiply()

        wavefront.planetype = 'image'

        return wavefront


class Detector(Image):

    def multiply(self, wavefront):
        # compute intensity
        #wavefront.data = [np.abs(w)**2 for w in wavefront.data]
        wavefront.planetype = 'image'
        return wavefront

    def frame(self, *args, **kwargs):
        """Simulate reading out a frame

        Returns
        -------
        frame : ndarray
            Raw frame

        Note
        ----
        This method just defines the ``frame`` interface. Any real functionality
        should be defined in a subclass.

        """
        raise NotImplementedError


class DispersivePhase(Plane):

    def multiply(self, wavefront):
        # NOTE: we can handle wavelength-dependent phase terms here (e.g. chromatic
        # aberrations). Since the phase will vary by wavelength, we can't fit out the
        # tilt pre-propagation and apply the same tilt for each wavelength like we can
        # with run of the mill tilt
        raise NotImplementedError


class DispersiveShift(Plane):

    def shift(self, wavelength, x0, y0, **kwargs):
        raise NotImplementedError

    def multiply(self, wavefront):
        wavefront = super().multiply(wavefront)
        wavefront.tilt.extend([self])
        return wavefront


class Grism(DispersiveShift):
    r"""Class for representing a grism.

    A grism is an optical element that can be inserted into a collimated beam
    to disperse incoming light according to its wavelength.

    Light is dispersed along a line called the the spectral trace. The position
    along the trace is determined by the dispersion function of the grism. The
    local origin of the spectral trace is anchored relative to the undispersed
    position of the source. This basic geometry is illustrated in the figure
    below:

    .. image:: /_static/img/grism_geometry.png
        :align: center
        :width: 400px

    The spectral trace is parameterized by a polynomial of the form

    .. math::

        y = a_n x^n + \cdots + a_2 x^2 + a_1 x + a_0

    and should return units of meters on the focal plane provided an input
    in meters on the focal plane.

    Similarly, the wavelength along the trace is parameterized by a
    polynomial of the form

    .. math::

        \lambda = a_n d^n + \cdots + a_2 d^2 + a_1 d + a_0

    and should return units of meters of wavelength provided an input distance
    along the spectral trace.

    Note
    ----
    Lentil supports trace and dispersion functions with any arbitrary polynomial
    order. While a simple analytic solution exists for modeling first-order trace
    and/or dispersion, there is no general solution for higher order functions.

    As a result, trace and/or dispersion polynomials with order > 1 are evaluated
    numerically. Although the effects are small, this approach impacts both the
    speed and precision of modeling grisms with higher order trace and/or
    dispersion functions. In cases where speed or accuracy are extremely important,
    a custom solution may be required.

    Parameters
    ----------
    trace : array_like
        Polynomial coefficients describing the spectral trace produced by the
        grism in decreasing powers (i.e. trace[0] represents the highest order
        coefficient and trace[-1] represents the lowest).

    dispersion : array_like
        Polynomial coefficients describing the dispersion produced by the grism
        in decreasing powers (i.e. dispersion[0] represents the highest order
        coefficient and dispersion[-1] represents the lowest.)

    See Also
    --------

    """
    def __init__(self, trace, dispersion, pixelscale=None, amplitude=1,
                 phase=0, mask=None, segmask=None):
        super().__init__(pixelscale=pixelscale, amplitude=amplitude, phase=phase,
                         mask=mask, segmask=segmask)

        self.trace = np.asarray(trace)
        self._trace_order = self.trace.size - 1
        assert self._trace_order >= 1

        self.dispersion = np.asarray(dispersion)
        self._dispersion_order = self.dispersion.size - 1
        assert self._dispersion_order >= 1

    def shift(self, wavelength, xs=0., ys=0., **kwargs):

        dist = self._dispersion(wavelength)
        x, y = self._trace(dist)

        # Include any incoming shift
        x += xs
        y += ys

        return x, y

    def _dispersion(self, wavelength):
        """Compute distance along spectral trace given wavelength.

        This method automatically selects the optimal computation stragegy
        depending on the dispersion model order.

        Parameters
        ----------
        wavelength: float
            Wavelength to compute dispersed distance in meters.

        Returns
        -------
        distance: float
            Distance along spectral trace relative to self.dispersion[-1]

        """

        if self._dispersion_order == 1:
            # For a first order polynomial we have lambda = dispersion[0] * dist + dispersion[1]
            # Solving for distance gives dist = (lambda - dispersion[1])/dispersion[0]
            return (wavelength - self.dispersion[1]) / self.dispersion[0]
        else:
            # Compute the arc length by numerically integrating from lambda_ref (dispersion[-1])
            # to wavelength
            #return self._arc_len(self._dispersion_dist_func, self.dispersion[-1], wavelength)
            return scipy.optimize.leastsq(self._dist_cost_func, x0=0, args=(wavelength,))[0]

    def _trace(self, dist):

        if self._trace_order == 1:
            # https://www.physicsforums.com/threads/formula-for-finding-point-on-a-line-given-distance-along-a-line.419561/
            x = dist/np.sqrt(1+self.trace[0]**2)
        else:
            # Find x by matching the observed distance along the trace computed by
            # self._arc_len(self._trace_dist_func(x), 0, x) with the known distance
            # along the trace dist as provided from the wavelength (via self._dispersion)
            x = scipy.optimize.leastsq(self._trace_cost_func, x0=0, args=(dist,))[0]

        y = np.polyval(self.trace, x)

        return x, y

    def _dispersion_wavelength_func(self, dist):
        # Integrand for computing wavelength (via numerical integration of arc length formula)
        # along the polynomial defined by coefficients in self.dispersion
        #return np.sqrt(1 + np.polyval(np.polyder(self.dispersion), dist)**2)
        return np.polyval(self.dispersion, dist)

    def _dist_cost_func(self, x, wavelength):
        return wavelength - self._dispersion_wavelength_func(x)

    def _trace_dist_func(self, x):
        # Integrand for computing distance along the trace (via numerical integration  of arc
        # length formula) along the polynomial defined by coefficients in self.trace
        return np.sqrt(1 + np.polyval(np.polyder(self.trace), x)**2)

    def _trace_cost_func(self, x, dist):
        # Compute difference between dist and distance computed along trace given x
        return dist - self._arc_len(self._trace_dist_func, 0, x)

    @staticmethod
    def _arc_len(dist_func, a, b):
        """Compute arc length by numerically evaluating the following expression for arc length:

        s = integrate sqrt(1+f'(x)^2) dx = integrate dist_func dx

        where dist_func is either _dispersion_dist_func or _trace_dist_func and f(x) is the
        polynomial defining either the trace or dispersion.

        Parameters
        ----------
        dist_func: func
            Function defining the integrand to be numerically integrated
        a, b: float
            Lower and upper integration bounds

        Reference
        ---------
        https://en.wikipedia.org/wiki/Arc_length#Finding_arc_lengths_by_integrating

        """
        return scipy.integrate.quad(dist_func, a, b)[0]


class LensletArray(Plane):
    pass


class Tilt(Plane):
    """Object for representing tilt in terms of an angle

    Parameters
    ----------
    x : float
        Radians of tilt about the x-axis

    y : float
        Radians of tilt about the y-axis

    """
    def __init__(self, x, y):
        super().__init__()

        self.x = y  # y tilt is about the x-axis.
        self.y = -x  # x tilt is about the y axis. There's also a sign flip to get the direction right.

    def multiply(self, wavefront):
        wavefront = super().multiply(wavefront)
        wavefront.tilt.extend([self])
        return wavefront

    def shift(self, xs=0, ys=0, z=0, **kwargs):
        """Compute image plane shift due to this angular tilt

        Parameters
        ----------
        xs : float
            Incoming x shift in meters. Default is 0.

        ys : float
            Incoming y shift in meters. Default is 0.

        z : float
            Propagation distance

        Returns
        -------
        shift : tuple
            Updated x and y shift terms

        """
        x = xs + (z * self.x)
        y = ys + (z * self.y)
        return x, y


class Rotate(Plane):
    """Rotate a Wavefront by a specified angle

    Parameters
    ----------
    angle : float
        Rotation angle counterclockwise from the horizontal.
    unit : {'degrees', 'radians'}, optional
        Units of angle. Default is 'degrees'.
    order : int
        The order of the spline interpolation (if needed), default is 3. The
        order has to be in the range 0-5.

    Note
    ----
    If the angle is an even multiple of 90 degrees, ``numpy.rot90`` is used to
    perform the rotation rather than ``scipy.ndimage.rotate``. In this case,
    the order parameter is irrelevant because no interpolation occurs.

    """
    def __init__(self, angle=0, unit='degrees', order=3):
        super().__init__()

        if unit == 'radians':
            angle *= 180/np.pi
        self.angle = -angle
        self.order = order

    def multiply(self, wavefront):
        """Multiply with a wavefront

        Parameters
        ----------
        wavefront : :class:`~lentil.wavefront.Wavefront` object
            Wavefront to be multiplied

        Returns
        -------
        wavefront : :class:`~lentil.wavefront.Wavefront` object
            Updated wavefront

        """
        if self.angle % 90 == 0:
            wavefront.data = np.rot90(wavefront.data, k=int(self.angle/90), axes=(1, 2))
        else:
            real = ndimage.rotate(wavefront.data.real, angle=self.angle, reshape=False, order=self.order, axes=(1, 2))
            imag = ndimage.rotate(wavefront.data.imag, angle=self.angle, reshape=False, order=self.order, axes=(1, 2))
            wavefront.data = real + 1j*imag
        return wavefront


class Flip(Plane):
    """Flip a wavefront along specified axis

    Parameters
    ----------
    axis : {int, tuple of ints, None}, optional
        Axis or axes along which to flip over. The default, axis=None, will
        flip over all of the axes of the input array. If axis is negative it
        counts from the last to the first axis.

    """

    def __init__(self, axis=None):
        super().__init__()

        if axis is None:
            # The first dimension of wavefront.data is depth so we actually want
            # to flip of the next two axes
            self.axis = (1, 2)
        else:
            # We convert axis to an array in case the user provides a tuple.
            # Again, because the first dimension of wavefront.data is depth we
            # add one to whatever axes are provided to make the flip operation
            # behave as expected.
            self.axis = np.asarray(axis) + 1

    def multiply(self, wavefront):
        """Multiply with a wavefront

        Parameters
        ----------
        wavefront : :class:`~lentil.wavefront.Wavefront` object
            Wavefront to be multiplied

        Returns
        -------
        wavefront : :class:`~lentil.wavefront.Wavefront` object
            Updated wavefront

        """
        wavefront.data = np.flip(wavefront.data, axis=self.axis)
        return wavefront


class Quadratic(Plane):
    """Base class for representing an optical plane with a quadratic phase
    term.

    """
    pass


class Conic(Plane):
    """Base class for representing an optical plane with a conic phase term.

    """
    pass


#    def Q(self, wave, pixelscale, oversample=1):
#        return (self.f_number*wave*oversample)/pixelscale
#
#    def alpha(self, wave, pixelscale, oversample=1):
#        return (self.pixelscale * pixelscale)/(self.focal_length * wave * oversample)
#
#    def q(self, wave, pixelscale, npix, oversample=1):
#        alpha = self.alpha(wave, pixelscale, oversample)
#        return 1/(alpha * npix)
